#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/base/commandline_application/commandline_arguments.rb'
# include Bioroebe::CommandlineArguments
# =========================================================================== #
module Bioroebe

module CommandlineArguments # === Bioroebe::CommandlineApplication

  # ========================================================================= #
  # === set_commandline_arguments
  #
  # This method can be used to set the commandline arguments, which
  # usually refers to ARGV.
  # ========================================================================= #
  def set_commandline_arguments(i = ARGV)
    i = [i].flatten.compact
    @commandline_arguments = i
  end

  # ========================================================================= #
  # === commandline_arguments?
  #
  # Query method to obtain all commandline arguments given to this class.
  # ========================================================================= #
  def commandline_arguments?
    @commandline_arguments
  end

  # ========================================================================= #
  # === first?
  # ========================================================================= #
  def first?
    @commandline_arguments.first
  end; alias first_argument? first? # === first_argument?
       alias first_argument  first? # === first_argument
       alias input_file?     first? # === input_file?

  # ========================================================================= #
  # === commandline_arguments_that_are_files?
  #
  # This method will select the entries that are a file (which must
  # exist locally).
  # ========================================================================= #
  def commandline_arguments_that_are_files?
    @commandline_arguments.select {|entry| File.file?(entry.to_s) }
  end

  # ========================================================================= #
  # === return_commandline_arguments_that_are_not_files
  #
  # This method will grab all entries that are NOT files.
  # ========================================================================= #
  def return_commandline_arguments_that_are_not_files
    @commandline_arguments.reject {|entry| File.file?(entry.to_s) }
  end

  # ========================================================================= #
  # === return_entries_without_two_leading_hyphens
  #
  # This variant will return all entries WITHOUT leading hyphens.
  # ========================================================================= #
  def return_entries_without_two_leading_hyphens(
      i = @commandline_arguments
    )
    i.reject {|entry| entry.start_with? '--' }
  end; alias return_commandline_arguments_without_hyphen return_entries_without_two_leading_hyphens # === return_commandline_arguments_without_hyphen

  # ========================================================================= #
  # === first_non_hyphen_argument?
  # ========================================================================= #
  def first_non_hyphen_argument?
    return_entries_without_two_leading_hyphens.first
  end

  # ========================================================================= #
  # === select_commandline_arguments
  #
  # This method can be used to select commandline arguments that begin
  # with -- hyphens.
  # ========================================================================= #
  def select_commandline_arguments(i = @commandline_arguments)
    i.select {|entry|
      entry.start_with? '--'
    }
  end

  # ========================================================================= #
  # === select_entries_starting_with_two_hyphens
  #
  # This entry can, specifically, be used to filter out proper
  # commandline arguments from an Array (such as ARGV).
  # ========================================================================= #
  def select_entries_starting_with_two_hyphens(
      i = commandline_arguments?
    )
    if i.respond_to? :select
      i.select {|entry| entry.start_with? '--' }
    else
      []
    end
  end; alias commandline_arguments_with_two_leading_hyphens         select_entries_starting_with_two_hyphens # === commandline_arguments_with_two_leading_hyphens
       alias commandline_arguments_containing_hyphens?              select_entries_starting_with_two_hyphens # === commandline_arguments_containing_hyphens?
       alias return_entries_with_two_leading_hyphens_from           select_entries_starting_with_two_hyphens # === return_entries_with_two_leading_hyphens_from
       alias return_arguments_with_leading_hyphens                  select_entries_starting_with_two_hyphens # === return_arguments_with_leading_hyphens
       alias return_commandline_arguments_starting_with_hyphens     select_entries_starting_with_two_hyphens # === return_commandline_arguments_starting_with_hyphens
       alias return_commandline_arguments_starting_with_two_hyphens select_entries_starting_with_two_hyphens # === return_commandline_arguments_starting_with_two_hyphens
       alias select_two_hyphens_from                                select_entries_starting_with_two_hyphens # === select_two_hyphens_from

  # ========================================================================= #
  # === remove_hyphens_from_the_commandline_arguments
  # ========================================================================= #
  def remove_hyphens_from_the_commandline_arguments(
      i = @commandline_arguments
    )
    i.reject! {|entry| entry.start_with? '--' }
  end; alias remove_hyphens_from remove_hyphens_from_the_commandline_arguments # === remove_hyphens_from

  # ========================================================================= #
  # === return_commandline_arguments_as_string
  #
  # This method will return the commandline arguments as a String.
  # ========================================================================= #
  def return_commandline_arguments_as_string
    @commandline_arguments.join(' ').strip
  end

end; end