#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
module Bioroebe

class CommandlineApplication < Base # === Bioroebe::CommandlineApplication

  require 'bioroebe/sequence/sequence.rb'
  require 'bioroebe/configuration/configuration.rb'
  require 'bioroebe/constants/files_and_directories.rb'
  require 'bioroebe/base/commandline_application/reset.rb'

  require 'bioroebe/codons/codons.rb'
  # ========================================================================= #
  # === stop_codons?
  # ========================================================================= #
  def stop_codons?(i)
    ::Bioroebe.stop_codons?(i)
  end

  require 'bioroebe/codons/codon_table.rb'
  # ========================================================================= #
  # === codon_table_dataset?
  # ========================================================================= #
  def codon_table_dataset?
    ::Bioroebe.codon_table_dataset?
  end

  require 'bioroebe/toplevel_methods/file_and_directory_related_actions.rb'
  # ========================================================================= #
  # === ensure_that_the_base_directory_exists
  #
  # Common functionality to ensure that the necessary base directories
  # exist on the target computer system.
  # ========================================================================= #
  def ensure_that_the_base_directories_exist
    ::Bioroebe.ensure_that_the_base_directories_exist
  end

  require 'bioroebe/toplevel_methods/aminoacids_and_proteins.rb'
  # ========================================================================= #
  # === strict_filter_away_invalid_aminoacids
  # ========================================================================= #
  def strict_filter_away_invalid_aminoacids(i)
    ::Bioroebe.strict_filter_away_invalid_aminoacids(i)
  end

  require 'bioroebe/toplevel_methods/leading_five_prime_and_trailing_three_prime.rb'
  # ========================================================================= #
  # === trailing_three_prime
  # ========================================================================= #
  def trailing_three_prime(
      i                 = '',
      get_rid_of_spaces = false,
      use_hyphen        = false
    )
    ::Bioroebe.trailing_three_prime(i, get_rid_of_spaces, use_hyphen)
  end

  require 'bioroebe/toplevel_methods/nucleotides.rb'
  # ========================================================================= #
  # === partner_nucleotide
  # ========================================================================= #
  def partner_nucleotide(i)
    ::Bioroebe.partner_nucleotide(i)
  end; alias return_dna_match partner_nucleotide # === return_dna_match

  require 'bioroebe/toplevel_methods/taxonomy.rb'
  # ========================================================================= #
  # === taxonomy_download_directory?
  # ========================================================================= #
  def taxonomy_download_directory?
    ::Bioroebe.taxonomy_download_directory?
  end

  # ========================================================================= #
  # === codons_for?
  #
  # See documentation for Bioroebe.codons_for? in the file
  # codons.rb, so it belongs to the above require statement.
  # ========================================================================= #
  def codons_for?(i)
    ::Bioroebe.codons_for?(i)
  end

  require 'bioroebe/toplevel_methods/chunked_display.rb'
  # ========================================================================= #
  # === return_chunked_display
  # ========================================================================= #
  def return_chunked_display(i)
    ::Bioroebe.chunked_display(i)
  end

  # ========================================================================= #
  # === return_cheerful_person
  # ========================================================================= #
  def return_cheerful_person
    '\\o/'
  end; alias cheerful_person return_cheerful_person # === cheerful_person

  # ========================================================================= #
  # === to_rna
  # ========================================================================= #
  def to_rna(i)
    i.tr('T','U')
  end

  require 'bioroebe/codons/start_codons.rb'
  # ========================================================================= #
  # === start_codon?
  # ========================================================================= #
  def start_codon?
    ::Bioroebe.start_codon?
  end

  # ========================================================================= #
  # === return_all_possible_start_codons
  #
  # This method will return all possible start codons. The code itself is
  # defined in the file "bioroebe/codons/start_codons.rb", so the
  # method here is grouped with the above .rb file (the require statement
  # above here).
  # ========================================================================= #
  def return_all_possible_start_codons
    ::Bioroebe.start_codons?
  end

  # ========================================================================= #
  # === is_this_a_start_codon?
  #
  # Note that the method-call defind in this method will return the start
  # codon as a DNA variant, so the given input should be in the form of
  # DNA rather than RNA.
  # ========================================================================= #
  def is_this_a_start_codon?(i)
    ::Bioroebe.start_codons?.include? i
  end

  # ========================================================================= #
  # === is_this_a_stop_codon?
  #
  # Query method to determine whether the given input is a stop codon or
  # whether it is not.
  # ========================================================================= #
  def is_this_a_stop_codon?(i)
    ::Bioroebe.stop_codons?.include? i
  end

  require 'bioroebe/toplevel_methods/cat.rb'
  # ========================================================================= #
  # === cat
  # ========================================================================= #
  def cat(i)
    ::Bioroebe.cat(i)
  end

  # ========================================================================= #
  # === codon_to_aminoacid
  # ========================================================================= #
  def codon_to_aminoacid(i)
    unless Bioroebe.respond_to? :codon_to_aminoacid
      require 'bioroebe/conversions/dna_to_aminoacid_sequence.rb'
    end
    ::Bioroebe.codon_to_aminoacid(i)
  end; alias nucleotides_to_aminoacid codon_to_aminoacid # === nucleotides_to_aminoacid
       alias codons_to_aminoacid      codon_to_aminoacid # === codons_to_aminoacid

  # ========================================================================= #
  # === without_extname
  #
  # This method will return a filename without the extname.
  # ========================================================================= #
  def without_extname(i)
    _ = File.extname(i)
    return i.gsub(/#{_}/, '')
  end

  # ========================================================================= #
  # === only_numbers?
  #
  # This method will return true if the input consists of only numbers.
  # Otherwise, false will be returned.
  # ========================================================================= #
  def only_numbers?(i)
    i = i.to_s
    i =~ /^\d+$/
  end

  # ========================================================================= #
  # === pad_with_double_quotes
  #
  # Pad the input with "" quotes.
  # ========================================================================= #
  def pad_with_double_quotes(i)
    return "\"#{i}\""
  end

  # ========================================================================= #
  # === pad_with_single_quotes
  #
  # Pad the input with '' quotes, unless it has them already.
  # ========================================================================= #
  def pad_with_single_quotes(i)
    _ = ''.dup
    _ << "'"
    _ << i.to_s.delete("'")
    _ << "'"
    return _
  end

  # ========================================================================= #
  # === load_bioroebe_yaml_file
  #
  # This method can be used to more easily load up a yaml file residing
  # in the main yaml/ directory of Bioroebe.
  #
  # If a Symbol is passed as input, then this will be assumed to correspond
  # to some existing yaml file there.
  # ========================================================================= #
  def load_bioroebe_yaml_file(
      i = :agarose
    )
    case i # case tag
    when :agarose
      i = 'agarose/agarose_concentrations.yml'
    end
    # ======================================================================= #
    # Aggregate it together with the BIOROEBE_YAML constant.
    # ======================================================================= #
    this_file = "#{PROJECT_YAML_DIRECTORY}"#{i}"
    if File.exist? this_file
      return YAML.load_file(this_file)
    else
      nil
    end
  end

  # ========================================================================= #
  # === one_letter_to_long_name
  #
  # The purpose of this method is really simple:
  #
  # - Take the one-letter abbreviation of an amino acid, and find
  #   the corresponding long name of the aminoacid.
  #
  # ========================================================================= #
  def one_letter_to_long_name(one_letter_code)
    inverted = AMINO_ACIDS_LONG_NAME_TO_ONE_LETTER.invert
    inverted[one_letter_code.upcase]
  end

  require 'bioroebe/toplevel_methods/editor.rb'
  # ========================================================================= #
  # === editor?
  #
  # Which editor we will use.
  # ========================================================================= #
  def editor?
    Bioroebe.editor?
  end

  # ========================================================================= #
  # === cp                                                 (cp tag, copy tag)
  #
  # Copy some files via this method here.
  # ========================================================================= #
  def cp(
      from, to = return_pwd
    )
    e "Now copying #{sfile(from.to_s)} to #{sfile(to.to_s)}."
    FileUtils.cp(from.to_s, to.to_s)
  end; alias copy cp # === copy

  # ========================================================================= #
  # === at_home?
  # ========================================================================= #
  def at_home?
    if ENV['IS_ROEBE']
      true
    else
      false
    end
  end; alias is_roebe? at_home? # === is_roebe?

  # ========================================================================= #
  # === remove_trailing_ansii_escape_code
  # ========================================================================= #
  def remove_trailing_ansii_escape_code(i)
    return ::Colours.remove_trailing_ansii_escape_code(i) if use_colours?
    i
  end

  # ========================================================================= #
  # === download_dir?
  # ========================================================================= #
  def download_dir?
    ::Bioroebe.download_dir?
  end

  require 'bioroebe/toplevel_methods/complement.rb'
  # ========================================================================= #
  # === complement                                           (complement tag)
  #
  # Usage example:
  #
  #   x = Bioroebe::Sequence.new('ATGGCA').complement # => "TACCGT"
  #
  # ========================================================================= #
  def complement(i)
    ::Bioroebe.complement(i)
  end; alias return_the_complementary_nucleotide_sequence_to complement # === return_the_complementary_nucleotide_sequence_to
       alias complementary_dna_strand                        complement # === complementary_dna_strand
       alias complementary_dna_sequence                      complement # === complementary_dna_sequence

  # ========================================================================= #
  # === esystem
  #
  # Run an external program - and output the result.
  # ========================================================================= #
  def esystem(i)
    ::Bioroebe.esystem(i)
  end

  # ========================================================================= #
  # === log_directory?
  # ========================================================================= #
  def log_directory?
    ::Bioroebe.log_directory?
  end; alias log_dir? log_directory? # === log_dir?

  # ========================================================================= #
  # === was_or_were
  # ========================================================================= #
  def was_or_were(i = 1)
    if i == 1
      'was'
    else
      'were'
    end
  end

  require 'bioroebe/toplevel_methods/return_ubiquitin_sequence.rb'
  # ========================================================================= #
  # === return_ubiquitin_sequence
  # ========================================================================= #
  def return_ubiquitin_sequence(i)
    ::Bioroebe.return_ubiquitin_sequence(i)
  end

  # ========================================================================= #
  # === three_to_one
  # ========================================================================= #
  def three_to_one(i)
    ::Bioroebe.three_to_one(i)
  end; alias three_aminoacid_letter_code_to_one_aminoacid_letter_code three_to_one # === three_aminoacid_letter_code_to_one_aminoacid_letter_code
       alias three_letters_to_one_letter                              three_to_one # === three_letters_to_one_letter

  # ========================================================================= #
  # === be_verbose?
  # ========================================================================= #
  def be_verbose?
    @be_verbose
  end; alias verbose? be_verbose? # === be_verbose?

  # ========================================================================= #
  # === be_silent
  # ========================================================================= #
  def be_silent
    @be_verbose = false
  end; alias be_quiet          be_silent # === be_quiet
       alias set_be_silent     be_silent # === set_be_silent
       alias set_be_quiet      be_silent # === set_be_quiet
       alias do_not_be_verbose be_silent # === do_not_be_verbose

  # ========================================================================= #
  # === leading_five_prime
  #
  # The first argument should be the input sequence; or, alternatively,
  # if you only need the "5' " header then you can omit it altogether.
  # ========================================================================= #
  def leading_five_prime(
      i                 = '',
      get_rid_of_spaces = false,
      use_hyphen        = false
    )
    ::Bioroebe.leading_five_prime(i, get_rid_of_spaces, use_hyphen)
  end

  # ========================================================================= #
  # === ccliner
  # ========================================================================= #
  def ccliner
    ee rev; cliner
  end

  require 'bioroebe/toplevel_methods/cliner.rb'
  # ========================================================================= #
  # === cliner
  # ========================================================================= #
  def cliner(
      use_this_token = :default_token,
      how_many_times = 80,
      &block
    )
    ::Bioroebe.cliner(
      use_this_token, how_many_times, &block
    )
  end

  # ========================================================================= #
  # === set_be_verbose
  # ========================================================================= #
  def set_be_verbose(i = true)
    case i
    when :be_verbose
      i = true
    when :be_quiet
      i = false
    end
    @be_verbose = i
  end

  # ========================================================================= #
  # === colourize_this_dna_sequence
  #
  # This method can be used to colourize a DNA sequence.
  # ========================================================================= #
  def colourize_this_dna_sequence(i)
    if i.is_a? ::Bioroebe::Sequence
      i = i.to_s
    end
    _ = ''.dup
    if use_colours?
      _ << remove_trailing_escape_code(
        ::Colours.send(USE_THIS_COLOUR_FOR_DNA)
      )
    end
    _ << i
    _ << rev if use_colours?
  end; alias colourize_dna_sequence colourize_this_dna_sequence # === colourize_dna_sequence
       alias colourize_nucleotides  colourize_this_dna_sequence # === colourize_nucleotides

  # ========================================================================= #
  # === remove_numbers
  #
  # This method will chop off numbers.
  # ========================================================================= #
  def remove_numbers(i)
    return i.delete('[0-9]')
  end

  require 'bioroebe/toplevel_methods/open_in_browser.rb'
  # ========================================================================= #
  # === open_in_browser
  # ========================================================================= #
  def open_in_browser(this_url)
    ::Bioroebe.open_in_browser(this_url)
  end

end; end