#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/calculate/calculate_levensthein_distance.rb'
# =========================================================================== #
module Bioroebe

  require 'bioroebe/colours/colours.rb'
  require 'bioroebe/colours/rev.rb'
  require 'bioroebe/toplevel_methods/e.rb'

  # ========================================================================= #
  # === Bioroebe.calculate_levensthein_distance
  #
  # The following method is based on
  #
  #   http://rosettacode.org/wiki/Levenshtein_distance#Ruby, slightly modified.
  #
  # To test this code, do:
  #
  #   [ ['kitten','sitting'], ['saturday','sunday'], ["rosettde", "raisethyrd"] ].each { |s,t|
  #     puts "calculate_levensthein_distance('#{s}', '#{t}') = #{Bioroebe.calculate_levensthein_distance(s, t)}"
  #   }
  #
  # However had, rubygems has a levensthein variant too.
  # ========================================================================= #
  def self.calculate_levensthein_distance(
      string1    = 'TTACCC',
      string2    = 'TTTCCC',
      be_verbose = true
    )
    case be_verbose
    when :be_quiet
      be_verbose = false
    end
    if string1.is_a?(Array) and (string1.size > 1)
      string2 = string1.shift
      string1 = string1.first
    elsif string1.is_a?(String) and string1.include?(' ')
      splitted = string1.split(' ')
      string2  = splitted.last
      string1  = splitted.first
    end
    m = string1.length
    n = string2.length
    return m if n == 0 # Stop at 0.
    return n if m == 0 # Stop at 0.
    arrays = Array.new(m+1) { Array.new(n+1) }
    # ======================================================================= #
    # Initialize the variable arrays next:
    # ======================================================================= #
    (0 .. m).each {|i| arrays[i][0] = i}
    (0 .. n).each {|j| arrays[0][j] = j}
    # ======================================================================= #
    # Now, iterate through:
    # ======================================================================= #
    (1 .. n).each {|j|
      (1 .. m).each {|i|
        arrays[i][j] = 
          if string1[i-1] == string2[j-1] # adjust index into string
            arrays[i-1][j-1]       # no operation required
          else
             [ arrays[i-1][j]+1,   # deletion     operation
               arrays[i][j-1]+1,   # insertion    operation
               arrays[i-1][j-1]+1, # substitution operation
             ].min
          end
      }
    }
    result = arrays[m][n]
    if be_verbose
      e rev+'The two strings '+simp(string1.to_s)+rev+' and '+
        simp(string2.to_s)+rev+' have n differences ('+
        steelblue('edit distance')+rev+'):'
      e "  #{simp(result.to_s)}"
    end
    return result
  end

end

if __FILE__ == $PROGRAM_NAME
  if ARGV.empty?
    alias e puts
    e
    e ::Bioroebe.rev+'Running in testing-mode '\
      'for the method Bioroebe.calculate_levensthein_distance():'
    e
    # Bioroebe.calculate_levensthein_distance(['saturday','sunday'])
    Bioroebe.calculate_levensthein_distance('CGTCAGCGTACTT','ATTC--CGTACTT')     # 1
    Bioroebe.calculate_levensthein_distance('CGGGGGGCTACTT','CGGTGGGT--GTG')     # 2
    Bioroebe.calculate_levensthein_distance('GAATCACGAAGC','GAA--ACGG-GC')       # 3
    Bioroebe.calculate_levensthein_distance('GCGATCAAGCCCGAG','GAG-TCGAGCCCA-G') # 4
    Bioroebe.calculate_levensthein_distance('CGGACCCAAGCTGAT','CTAA-CCA-GCTGAG') # 5
    Bioroebe.calculate_levensthein_distance('ATA','A-A')                         # 6
  else
    Bioroebe.calculate_levensthein_distance(*ARGV)
  end
end # calculatelevenstheindistance