#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::ParseFrequencyTable
#
# This class can be used to "convert" a .cgi entry such as
# https://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=9685
# into a corresponding .yml file.
#
# This then allows us to use that species.
#
# Usage example:
#
#   Bioroebe::ParseFrequencyTable.new
#
# =========================================================================== #
# require 'bioroebe/codon_tables/frequencies/parse_frequency_table.rb'
# Bioroebe::ParseFrequencyTable
# =========================================================================== #
require 'bioroebe/base/commandline_application/commandline_application.rb'

module Bioroebe

class ParseFrequencyTable < ::Bioroebe::CommandlineApplication # === Bioroebe::ParseFrequencyTable

  require 'open-uri'

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === AUTOMATICALLY_CREATE_A_LOCAL_YAML_FILE_IF_IT_DOES_NOT_YET_EXIST
  #
  # If this constant is set to true then this class will attempt to
  # create a local .yml file, unless such a file already exists.
  #
  # That .yml file will include the keys and frequency values associated
  # with these keys, such as: "UUU: 17.7".
  #
  # The idea behind this is so that we can simplify the creation of such
  # .yml files, which in turn may then be used to populate a SQL
  # database.
  # ========================================================================= #
  AUTOMATICALLY_CREATE_A_LOCAL_YAML_FILE_IF_IT_DOES_NOT_YET_EXIST = true

  # ========================================================================= #
  # === DEFAULT_INPUT
  #
  # This is just sample-data that can be used to test the inner-working
  # of this class.
  # ========================================================================= #
  DEFAULT_INPUT = 'UUU 17.7(  2478)  UCU 14.2(  1994)  UAU 12.1(  1698)  UGU 10.5(  1473)
UUC 25.9(  3630)  UCC 18.9(  2641)  UAC 18.9(  2640)  UGC 13.8(  1928)
UUA  6.3(   880)  UCA  9.5(  1328)  UAA  0.7(   102)  UGA  1.5(   216)
UUG 13.5(  1892)  UCG  4.9(   684)  UAG  0.5(    64)  UGG 15.3(  2148)

CUU 11.9(  1667)  CCU 13.9(  1947)  CAU  8.7(  1219)  CGU  3.7(   519)
CUC 22.9(  3210)  CCC 20.4(  2860)  CAC 14.8(  2076)  CGC  9.8(  1371)
CUA  6.4(   897)  CCA 13.1(  1838)  CAA 11.6(  1630)  CGA  5.0(   696)
CUG 45.8(  6407)  CCG  6.8(   952)  CAG 30.1(  4212)  CGG 10.7(  1502)

AUU 14.9(  2081)  ACU 11.4(  1594)  AAU 17.0(  2378)  AGU 10.2(  1426)
AUC 24.9(  3483)  ACC 21.6(  3020)  AAC 24.2(  3391)  AGC 18.8(  2630)
AUA  7.4(  1041)  ACA 12.7(  1784)  AAA 23.3(  3255)  AGA 12.0(  1678)
AUG 22.4(  3140)  ACG  8.5(  1186)  AAG 30.7(  4292)  AGG 12.7(  1779)

GUU 10.2(  1421)  GCU 16.5(  2314)  GAU 19.5(  2733)  GGU  9.2(  1293)
GUC 18.3(  2566)  GCC 31.4(  4391)  GAC 26.8(  3758)  GGC 22.5(  3143)
GUA  6.2(   869)  GCA 13.1(  1833)  GAA 27.1(  3800)  GGA 16.0(  2239)
GUG 30.5(  4267)  GCG  8.6(  1207)  GAG 35.8(  5005)  GGG 15.6(  2181)'

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = ARGV,
      run_already = true
    )
    reset
    set_commandline_arguments(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @use_this_url
    #
    # This variable will be nil on startup, so that this class can decide
    # whether to download from a remote URL or not.
    # ======================================================================= #
    @use_this_url = nil
    # ======================================================================= #
    # === @hash_storing_the_codon_frequencies
    # ======================================================================= #
    @hash_storing_the_codon_frequencies = {}
  end

  # ========================================================================= #
  # === set_commandline_arguments
  # ========================================================================= #
  def set_commandline_arguments(i)
    @commandline_arguments = [i].flatten.compact
    menu(@commandline_arguments)
  end

  # ========================================================================= #
  # === menu                                                       (menu tag)
  # ========================================================================= #
  def menu(
      i = @commandline_arguments
    )
    if i.is_a? Array
      i.each {|entry| menu(entry) }
    else
      case i # case tag
      # ===================================================================== #
      # === --random
      #
      # This entry point will try to randomly grab an (any) entry.
      # Thus it will often fail.
      # ===================================================================== #
      when /^-?-?random$/
        i = return_url_from_a_random_entry
        set_url(i)
      end
    end
  end

  # ========================================================================= #
  # === return_url_from_a_random_entry                           (random tag)
  # ========================================================================= #
  def return_url_from_a_random_entry
    'https://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species='+
    (rand(100_000)+1).to_s
  end

  # ========================================================================= #
  # === first?
  # ========================================================================= #
  def first?
    @commandline_arguments.first
  end; alias input?            first? # === input?
       alias original_dataset? first? # === original_dataset?

  # ========================================================================= #
  # === set_first
  # ========================================================================= #
  def set_first(i)
    @commandline_arguments[0] = i
  end

  # ========================================================================= #
  # === set_use_this_url
  # ========================================================================= #
  def set_use_this_url(i = nil)
    if i.is_a? Array
      i = i.first
    end
    i = SAMPLE_INPUT if i.nil? # Use default in this case.
    i = i.to_s.dup
    @use_this_url = i
  end; alias set_url set_use_this_url # === set_url

  # ========================================================================= #
  # === use_this_url?
  # ========================================================================= #
  def use_this_url?
    @use_this_url
  end; alias remote_url? use_this_url? # === remote_url?
       alias url?        use_this_url? # === url?

  # ========================================================================= #
  # === determine_whether_we_have_to_download_from_a_remote_url
  #
  # This method will determine whether this class has to download the
  # dataset from a remote URL.
  # ========================================================================= #
  def determine_whether_we_have_to_download_from_a_remote_url
    first = first?
    if first and first.start_with?('http')
      set_use_this_url(first)
    end
  end

  # ========================================================================= #
  # === report_the_frequencies_properly_formatted
  # ========================================================================= #
  def report_the_frequencies_properly_formatted
    @hash_storing_the_codon_frequencies.each_pair {|key, value|
      e "#{key}: #{value.to_s.rjust(4)}"
    }
  end

  # ========================================================================= #
  # === store_locally?
  # ========================================================================= #
  def store_locally?
    AUTOMATICALLY_CREATE_A_LOCAL_YAML_FILE_IF_IT_DOES_NOT_YET_EXIST
  end

  # ========================================================================= #
  # === do_store_into_a_local_yaml_file
  # ========================================================================= #
  def do_store_into_a_local_yaml_file(
      into = :default
    )
    case into
    when :default
      into = 'default_yaml_file.yml'
    end
    if File.exist? into
      e "Can not store into the local file #{sfile(into)} as such a file "\
        "already exists."
    else
      remote_url = remote_url?
      what = ''.dup
      if remote_url.include? '?species='
        # =================================================================== #
        # Assume an URL such as this one here:
        #
        #   https://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=107243
        #
        # =================================================================== #
        if into.include? 'default_yaml'
          into = remote_url.split('=').last.strip
          # ================================================================= #
          # We can also use a regex on the original dataset to obtain
          # the name of the organism, such as here:
          #
          #   <STRONG><i>Thlaspi caerulescens </i>[gbpln]: 45 CDS's (20416 codons)</STRONG>
          #
          # ================================================================= #
          original_dataset = original_dataset?
          if original_dataset.include? '<STRONG>'
            use_this_regex = /<STRONG><i>(.+) <\/i>/ # See: https://rubular.com/r/o4WJAeRxuvB4kM
            original_dataset =~ use_this_regex
            match = $1.to_s.dup.strip.tr(' ','_')
            into << "_#{match}"
          end
          into << '.yml' unless into.end_with? '.yml'
        end
      end
      if @hash_storing_the_codon_frequencies.empty?
        e 'No dataset was found at the URL `'+sfancy(url?)+'`.'
      else
        # =================================================================== #
        # First create a generic header for that file:
        # =================================================================== #
        what << "#\n"
        what << "# Filename: #{into}\n"
        what << "#\n"
        what << "# The following dataset was obtained from:\n"
        what << "#\n"
        what << "#   #{remote_url}\n"
        what << "#\n"
        what << "#\n"
        @hash_storing_the_codon_frequencies.each_pair {|key, value|
          what << "#{key}: #{value.to_s.rjust(4)}\n"
        }
        e "Saving into the local file at `#{sfile(into)}`."
        write_what_into(what, into)
      end
    end
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    determine_whether_we_have_to_download_from_a_remote_url
    # ======================================================================= #
    # See here:
    #
    #   https://rubular.com/r/NgSPXJ2f5WG2kq
    #
    # ======================================================================= #
    if @use_this_url
      use_this_regex = /([A-Z]{3})  ?(\d{1,2}\.\d{0,2})/
      opnn; e 'Obtaining dataset from the following remote URL:'
      e
      e sfancy("  #{@use_this_url}")
      e
      _ = @use_this_url
      set_first(
        URI.open(_).read
      )
    end
    splitted = first?.split("\n").reject {|line| line.to_s.strip.empty? }
    splitted.each {|line|
      # ===================================================================== #
      # This line may look like:
      #
      #   UUC 25.9(  3630)  UCC 18.9(  2641)  UAC 18.9(  2640)  UGC 13.8(  1928)
      #
      # ===================================================================== #
      inner_splitted = line.scan(use_this_regex)
      inner_splitted.each {|key, value|
        @hash_storing_the_codon_frequencies[key] = value
      }
    }
    report_the_frequencies_properly_formatted
    do_store_into_a_local_yaml_file if store_locally?
  end

  # ========================================================================= #
  # === Bioroebe::ParseFrequencyTable[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

end; end

if __FILE__ == $PROGRAM_NAME
  Bioroebe::ParseFrequencyTable.new(ARGV)
end # parsefrequencytable https://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=9685
    # parsefrequencytable https://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=3702
    # parsefrequencytable https://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=107243
    # parsefrequencytable --random