#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# = Description
#
# Popular colour codings for nucleic and amino acids.
# 
# The Bioroebe::ColourScheme module contains classes that return popular
# colour codings for nucleic and amino acids in RGB hex format suitable
# for HTML code.
#
# Presently the following schemes are supported:
#
#     Buried     - Buried index
#     Helix      - Helix propensity
#     Hydropathy - Hydrophobicity
#     Nucleotide - Nucelotide colour coding
#     Strand     - Strand propensity
#     Taylor     - Taylor colour coding
#     Turn       - Turn propensity
#     Zappo      - Zappo colour coding
# 
# More possible colour schemes may be these variants:
#
#   BLOSUM62
#   ClustalX
#   Percentage Identity (PID)
#
# The colours schemes BLOSUM62, ClustalX, and Percentage Identity
# are all dependent on the alignment consensus.
#
# This data is currently referenced from the "JalView alignment editor".
#
#   Clamp, M., Cuff, J., Searle, S. M. and Barton, G. J. (2004), 
#   "The Jalview Java Alignment Editor," Bioinformatics, 12, 426-7
#
#     http://www.jalview.org
# 
# Currently, the score data for things such as hydropathy, helix, turn,
# etc. are contained here but should probably be moved elsewhere once a
# good reference is found for these values.
#
# = Usage example
# 
#     require 'bioroebe'
# 
#     seq = 'gattaca'
#     scheme = Bioroebe::ColourScheme::Zappo
#     postfix = '</span>'
#     html = ''.dup
#     seq.each_byte { |c|
#       colour = scheme[c.chr]
#       prefix = %Q(<span style="background:\##{colour};">)
#       html << prefix + c.chr + postfix
#     }
#     puts html
#
# == Accessing colours
# 
#   puts Bioroebe::ColourScheme::Buried['A']    # 00DC22
#   puts Bioroebe::ColourScheme::Buried[:c]     # 00BF3F
#   puts Bioroebe::ColourScheme::Buried[nil]    # nil
#   puts Bioroebe::ColourScheme::Buried['-']    # FFFFFF
#   puts Bioroebe::ColourScheme::Buried[7]      # FFFFFF
#   puts Bioroebe::ColourScheme::Buried['junk'] # FFFFFF
#   puts Bioroebe::ColourScheme::Buried['t']    # 00CC32
# 
# =========================================================================== #
# External Resources that may be useful here:
#
#   https://www.jalview.org/help.html
#
# =========================================================================== #
# require 'bioroebe/colour_schemes/colour_scheme.rb'
# =========================================================================== #
require 'bioroebe/colours/colour_schemes/simple.rb'
require 'bioroebe/colours/colour_schemes/score.rb'

module Bioroebe

module ColourScheme # === Bioroebe::ColourScheme

  # ========================================================================= #
  # === Consensus sub-classes
  #
  # NOTE todo:
  #
  #   BLOSUM62
  #   ClustalX
  #   PID
  #
  # ========================================================================= #

  # ========================================================================= #
  # TODO - anyone wants to implement these?
  # ========================================================================= #
  class Consensus
  end

end; end