#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::ColourSchemeDemo
#
# A Bioroebe::ColourScheme demo script for Nucleic Acids sequences.
#
# Usage example:
#
# (1) If you want to showcase what this class can do, run a command
#     similar to the following one:
#
#       colour_scheme > /Depot/jjj/foo.html; open /Depot/jjj/foo.html
#
#     Where colour_scheme is an alias to this file here.
#
# (2) If you want to colourize sequences, try this:
#
#       cat sequence.fasta
#
#       >DNA_sequence
#       acgtgtgtcatgctagtcgatcgtactagtcgtagctagtca
#
#       colour_scheme sequence.fasta > colourized_sequence.html; open colourized_sequence.html
# 
# =========================================================================== #
# require 'bioroebe/colour_schemes/colour_scheme_demo.rb'
# =========================================================================== #
require 'bioroebe/base/base.rb'
require 'bioroebe/colours/colour_schemes/colour_scheme.rb'
require 'bioroebe/colours/colour_schemes/array_available_colour_schemes.rb'

Bioroebe::ColourScheme::ARRAY_AVAILABLE_COLOUR_SCHEMES.each {|this_scheme|
  require "bioroebe/colours/colour_schemes/#{this_scheme}.rb"
}

module Bioroebe

class ColourSchemeDemo < ::Bioroebe::Base

  require 'bioroebe/conversions/dna_to_aminoacid_sequence.rb'
  require 'bioroebe/fasta_and_fastq/parse_fasta/parse_fasta.rb'
  require 'bioroebe/toplevel_methods/nucleotides.rb'

  # ========================================================================= #
  # === TITLE
  # ========================================================================= #
  TITLE = 'Bioroebe::ColourScheme for DNA sequences'

  # ========================================================================= #
  # === CDIV
  # ========================================================================= #
  CDIV = '</div>'

  # ========================================================================= #
  # === BODY_TAG
  # ========================================================================= #
  BODY_TAG = '<body style="font-size:1.5em">'

  # ========================================================================= #
  # === MAXIMAL_N_NUCLEOTIDES_PER_ROW
  # ========================================================================= #
  MAXIMAL_N_NUCLEOTIDES_PER_ROW = 80

  # ========================================================================= #
  # === DEFAULT_FASTA_FILE_NAME  
  # ========================================================================= #
  DEFAULT_FASTA_FILE_NAME = 'tardigrada_fasta.ffn'

  # ========================================================================= #
  # === initialize
  #
  # First argument should be the nucleotide sequence.
  #
  # Second argument should be the aminoacid sequence. But it can be
  # omitted - in that case, we will deduce the aminoacid sequence
  # from the nucleotide sequence.
  # ========================================================================= #
  def initialize(
      nucleotide_sequence,
      aminoacid_sequence = nil
    )
    reset
    case aminoacid_sequence
    when :be_quiet
      aminoacid_sequence = nil
      @be_verbose = false
    end
    # ======================================================================= #
    # === Handle the case when the user did not input an aminoacid sequence
    # ======================================================================= #
    if aminoacid_sequence.nil?
      if nucleotide_sequence.respond_to? :translate
        aminoacid_sequence = nucleotide_sequence.translate # Aminoacid sequence.
      elsif nucleotide_sequence.is_a? String
        aminoacid_sequence = ::Bioroebe::DnaToAminoacidSequence.new(
          nucleotide_sequence, :return_silently
        )
      end
    end
    _ = ['<html>',
         '<header>', '<title>', TITLE, '</title>', '</header>',
         BODY_TAG,  '<h1>', TITLE, '</h1>']
    _ << ['<div>', '<h2>', 'Simple colors', '</h2>']
    %w( Nucleotide ).each { |scheme|
      _ << display_scheme(scheme, nucleotide_sequence, '')
    }
    _ << [CDIV]
    %w( Zappo Taylor ).each { |scheme|
      _ << display_scheme(scheme, '', aminoacid_sequence)
    }
    _ << [CDIV]
    _ << ['<div>', '<h2>', 'Score colors', '</h2>']
    %w( Buried Helix Hydropathy Strand Turn ).each { |score|
      _ << display_scheme(score, "", aminoacid_sequence)
    }
    _ << [CDIV]+['</body>','</html>']
    @result = _
    report_result if @be_verbose
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @be_verbose
    # ======================================================================= #
    @be_verbose = true
  end

  # ========================================================================= #
  # === br
  #
  # Returns folded sequence with <br> HTML tag.
  #
  # The second argument keeps track when to add a <br> tag and a newline.
  # ========================================================================= #
  def br(
      i, width = MAXIMAL_N_NUCLEOTIDES_PER_ROW
    )
    return "<br\n>" if i % width == 0
    return '' # Else return an empty string.
  end

  # ========================================================================= #
  # === display_scheme
  #
  # This method will properly show the sequence. It will use a <div>
  # tag for this.
  #
  # Returns scheme wise html doc.
  # ========================================================================= #
  def display_scheme(
      scheme,
      nucleotide_sequence,
      aminoacid_sequence
    )
    html = ''.dup
    # ======================================================================= #
    # === Obtain the relevant constant next
    # ======================================================================= #
    constant = ::Bioroebe::ColourScheme.const_get(scheme)
    [ nucleotide_sequence, aminoacid_sequence ].each { |sequence|
      html << display(sequence, constant)
    }
    result = ['<div>', "<h3>#{constant}</h3>", html, CDIV]
    return result
  end

  # ========================================================================= #
  # === display
  #
  # This method will return the sequence html doc, which equates to
  # a <p> tag in HTML.
  # ========================================================================= #
  def display(sequence, cs)
    # ======================================================================= #
    # Open the <p> tag first:
    # ======================================================================= #
    html = '<p style="font-family: monospace; padding-left: 1.25em;">'.dup
    close_span_tag = '</span>'
    index = 0 # Index.
    if sequence.is_a? ::Bioroebe::DnaToAminoacidSequence
      sequence = sequence.sequence? # We need the String here.
    end
    sequence.each_byte {|char|
      colour = cs[char.chr]
      open_span_tag = '<span style="background:#'+colour+';">'
      html << open_span_tag+char.chr.to_s+close_span_tag
      html << br(index += 1) # Increment the index too.
    }
    html+'</p>'
  end

  # ========================================================================= #
  # === result?
  # ========================================================================= #
  def result?
    _ = @result
    _ = _.join("\n") if _.is_a? Array
    return _
  end

  # ========================================================================= #
  # === report_result
  # ========================================================================= #
  def report_result
    e result? # Call it by default.
  end

  # ========================================================================= #
  # === Bioroebe::ColourSchemeDemo.create_demo_file
  #
  # The following class method can be used to create a demo file.
  #
  # It assumes that a fasta file will exist at the base directory.
  # ========================================================================= #
  def self.create_demo_file(
      at_this_location = "#{log_dir?}#{DEFAULT_FASTA_FILE_NAME}"
    )
    # ======================================================================= #
    # === We must first check whether the file in question exists
    # ======================================================================= #
    if File.exist? at_this_location
      sequence = ::Bioroebe.parse_fasta_file(at_this_location).sequence?
      what = self.new(sequence, :be_quiet).result?
      into = '/Depot/j/colourized_sequence.html'
      write_what_into(what, into)
      return into
    else
      puts 'No file could be found at `'+
            ::Colours.sfile(at_this_location)+rev+'`.'
      return nil
    end
  end

  # ========================================================================= #
  # === Bioroebe::ColourSchemeDemo.rev
  # ========================================================================= #
  def self.rev
    ::Bioroebe.rev
  end

end; end

if __FILE__ == $PROGRAM_NAME
  # =========================================================================== #
  # Must be ported into Bioroebe eventually.
  # =========================================================================== #
  file = ARGV.shift
  if file # If an argument was suuplied.
    nucleotide_sequence = Bioroebe.parse_fasta_file(file).sequence?
  else # Else use a random sequence.
    nucleotide_sequence = Bioroebe.create_random_dna_sequence(250)
  end
  Bioroebe::ColourSchemeDemo.new(nucleotide_sequence)
end # cscheme > /Depot/j/foo.html
    # cscheme /root/bioroebe/NM_001180897.3_Saccharomyces_cerevisiae_S288c_Aga2p_AGA2.fasta > /Depot/j/colourized_sequence.html; open /Depot/j/colourized_sequence.html
    # cscheme /root/Bioroebe/1fat.fasta > /Depot/j/foo.html; oib /Depot/j/foo.html