#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::ColourScheme::Score
#
# A class template for colour code referencing of colour schemes that
# are score based. This template is expected to change when the scores
# are moved into bioroebe/data/aa.
# =========================================================================== #
module Bioroebe

module ColourScheme

class Score # === Bioroebe::ColourScheme::Score

  # ========================================================================= #
  # === @colours
  #
  # Class instance variables.
  # ========================================================================= #
  @colours = {}
  @colours.default = 'FFFFFF' # Use white colour by default.
  @scores = {
    'A' => 0.83,
  }
  @min = 0.37
  @max = 1.7

  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score.percent_to_hex
  # ========================================================================= #
  def self.percent_to_hex(percent)
    percent = percent.to_f if percent.is_a? String
    if (percent > 1.0) or (percent < 0.0) or percent.nil?
      raise 'Percentage must be between 0.0 and 1.0'
    end
    '%02X' % (percent * 255.0)
  end

  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score.rgb_percent_to_hex
  # ========================================================================= #
  def self.rgb_percent_to_hex(red, green, blue)
    Score.percent_to_hex(red)+   # R
    Score.percent_to_hex(green)+ # G
    Score.percent_to_hex(blue)   # B
  end

  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score.score_to_rgb_hex
  # ========================================================================= #
  def self.score_to_rgb_hex(score, min, max)
    percent = score_to_percent(score, min, max)
    Score.rgb_percent_to_hex(percent, 0.0, 1.0-percent)
  end
  
  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score.score_to_percent
  # ========================================================================= #
  def self.score_to_percent(score, min, max)
    # .to_f to ensure every operation is float-aware.
    percent = (score.to_f - min) / (max.to_f - min)
    percent = 1.0 if percent > 1.0
    percent = 0.0 if percent < 0.0
    percent
  end

  # ========================================================================= #
  # === Populate @scores next
  # ========================================================================= #
  @scores.each { |key, value|
    @colours[key] = Score.score_to_rgb_hex(value, @min, @max)
  }

  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score.min
  # ========================================================================= #
  def self.min(i = nil)
    @min
  end

  # ========================================================================= #
  # === Score.max
  # ========================================================================= #
  def self.max(i = nil)
    @max
  end

  # ========================================================================= #
  # === Score.scores
  # ========================================================================= #
  def self.scores
    @scores
  end

  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score.colours
  # ========================================================================= #
  def self.colours
    @colours
  end; self.instance_eval { alias colors colours } # === Bioroebe::ColourScheme::Score.colors

  # ========================================================================= #
  # === Bioroebe::ColourScheme::Score[]
  # ========================================================================= #
  def self.[](i)
    return if i.nil?
    @colours[i.to_s.upcase]
  end

end; end; end