#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/constants/nucleotides.rb'
# =========================================================================== #
module Bioroebe

  # ========================================================================= #
  # === Bioroebe::STOP_CODONS
  #
  # The STOP codons that can be found in Humans, in RNA format.
  # ========================================================================= #
  STOP_CODONS = %w(
    UAA UAG UGA
  )

  # ========================================================================= #
  # === Bioroebe::RNA_NUCLEOTIDES
  #
  # This will refer to an Array including all four RNA nucleotides,
  # that is A, U, G and C.
  # ========================================================================= #
  RNA_NUCLEOTIDES = %w( A U G C )
    ALLOWED_RNA_NUCLEOTIDES = RNA_NUCLEOTIDES # === ALLOWED_RNA_NUCLEOTIDES

  # ========================================================================= #
  # === Bioroebe::POSSIBLE_RNA_NUCLEOTIDES
  #
  # This is a bit different to RNA_NUCLEOTIDES in that N is also a part
  # of it. It is not entirely clear whether this array here is kept,
  # though.
  # ========================================================================= #
  POSSIBLE_RNA_NUCLEOTIDES = %w(
    A U G C N
  )

  # ========================================================================= #
  # === Bioroebe::DNA_NUCLEOTIDES
  #
  # This is the variant without N.
  # ========================================================================= #
  DNA_NUCLEOTIDES = %w( A T G C )

  # ========================================================================= #
  # === Bioroebe::HASH_DNA_NUCLEOTIDES
  #
  # Since as of 20.04.2014, Uracil is also part of this Hash. While this
  # is, strictly speaking, not absolutely correct, it does simplify some
  # downstream code. However had, this may possibly be re-evaluated in 
  # the future.
  #
  # This Hash may be helpful when the user wishes to find a complement to
  # a nucleotide. There is a method that does the same, but this Hash
  # should be faster than a method call, so use it in particular if 
  # you need to focus more on speed.
  # ========================================================================= #
  HASH_DNA_NUCLEOTIDES = {
    'A' => 'T',
    'T' => 'A',
    'G' => 'C',
    'C' => 'G',
    'U' => 'A'
  }

  # ========================================================================= #
  # === Bioroebe.this_nucleotide_pairs_with?
  # ========================================================================= #
  def self.this_nucleotide_pairs_with?(i)
    HASH_DNA_NUCLEOTIDES[i.to_s.upcase]
  end

  # ========================================================================= #
  # === POSSIBLE_DNA_NUCLEOTIDES
  #
  # This constant will keep all possible DNA nucleotides.
  #
  # N is also a valid entry, 'Yarrowia_lipolytica_genome.fa' includes it.
  # However had, 
  #
  # Only these sequences are allowed in DNA.
  #
  # To scope to this, do:
  #
  #   Bioroebe::POSSIBLE_DNA_NUCLEOTIDES
  #
  # ========================================================================= #
  POSSIBLE_DNA_NUCLEOTIDES = %w(
    A T G C N
  ); ARRAY_VALID_DNA_SEQUENCES = POSSIBLE_DNA_NUCLEOTIDES # === ARRAY_VALID_DNA_SEQUENCES

  # ========================================================================= #
  # === Bioroebe.return_random_nucleotide
  #
  # This method will randomly return either 'A', 'T', 'C' or 'G'.
  # ========================================================================= #
  def self.return_random_nucleotide
    (POSSIBLE_DNA_NUCLEOTIDES - ['N']).sample # This constant holds A T G and C.
  end

  # ========================================================================= #
  # === Bioroebe.allowed_dna_nucleotides?
  #
  # This will return an Array with valid DNA nucleotides.
  # ========================================================================= #
  def self.allowed_dna_nucleotides?
    ALLOWED_DNA_NUCLEOTIDES
  end; self.instance_eval { alias dna? allowed_dna_nucleotides? } # === Bioroebe.dna?

end

if __FILE__ == $PROGRAM_NAME
  pp Bioroebe.allowed_dna_nucleotides?
end