#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::Palindrome2DStructure
#
# This class will draw a RNA palindrome or DNA palindrome in "2D".
#
# Note that the class is currently limited in that it can not handle
# non-pairing nucleotides. 
# =========================================================================== #
# require 'bioroebe/palindromes/palindrome_2D_structure.rb'
# Bioroebe::Palindrome2DStructure.new(ARGV)
# =========================================================================== #
require 'bioroebe/base/commandline_application/commandline_application.rb'

module Bioroebe

class Palindrome2DStructure < ::Bioroebe::CommandlineApplication # === Bioroebe::Palindrome2DStructure

  # ========================================================================= #
  # === DEFAULT_SEARCH_STRING
  # ========================================================================= #
  DEFAULT_SEARCH_STRING = 'CGCGAATTCGCG'

  # ========================================================================= #
  # === initialize
  #
  # Can also invoke it via rpalindrome.
  # Currently we can find only palindromes which have odd numbers
  # ========================================================================= #
  def initialize(
      i           = ARGV,
      run_already = true
    )
    reset
    set_commandline_arguments(i)
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
  end

  # ========================================================================= #
  # === determine_the_main_sequence
  # ========================================================================= #
  def determine_the_main_sequence
    set_sequence(first?)
  end
  
  # ========================================================================= #
  # === set_sequence
  # ========================================================================= #
  def set_sequence(i = first?)
    if i
      i = i.dup if i.frozen?
      i.delete!('|')
    end
    @sequence = i
  end

  # ========================================================================= #
  # === sequence?
  # ========================================================================= #
  def sequence?
    @sequence
  end

  # ========================================================================= #
  # === display_these_two_nucleotides
  # ========================================================================= #
  def display_these_two_nucleotides(a, b)
    case a # case tag
    when 'A','T','U'
      use_this_token = '='
      n_hydrogen_bonds = 2
    when 'G','C'
      use_this_token = '≡'
      n_hydrogen_bonds = 3
    end
    unless can_base_pair?(a, b)
      use_this_token = ''
      n_hydrogen_bonds = 0
    end
    e "     #{a}#{use_this_token}#{b}  # #{n_hydrogen_bonds}"
  end

  # ========================================================================= #
  # === display_trailing_connecters
  # ========================================================================= #
  def display_trailing_connecters
    e '    /   \\'
  end

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    determine_the_main_sequence
    _ = sequence?
    if is_palindrome? _
      # ===================================================================== #
      # We only support palindromes for the time being. If the sequence
      # is a palindrome, in the DNA or RNA, then we know that it must
      # have an even amount of nucleotides.
      # ===================================================================== #
      size = _.size
      half = size / 2
      # ===================================================================== #
      # To display it in "2D", we start with the two top nucleotides,
      # that is those that are found first when looking at the structure,
      # then we continue. 
      # ===================================================================== #
      e
      half.times {|n_time|
        a = _[half-(n_time+1)] # The first nucleotide.
        b = _[half+n_time]     # The complementary nucleotide.
        display_these_two_nucleotides(a, b)
      }
      display_trailing_connecters
      e
    else
      erev 'The sequence is not a palindrome. Currently this class can only'
      erev 'display palindromes on the commandline.'
    end
  end

end; end

if __FILE__ == $PROGRAM_NAME
  Bioroebe::Palindrome2DStructure.new(ARGV)
end # palindrome2dstructure CGCGAATTCGCG
    # palindrome2dstructure "CGCGAA|TTCGCG"
    # palindrome2dstructure CGCGAATTCGCGGCGCTTAAGCGC
    # palindrome2dstructure CGCG