#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::PalindromeGenerator
#
# This class will generate a palindromic sequence.
#
# The first input is the length.
#
# Usage example:
#
#   Bioroebe::PalindromeGenerator.new(ARGV)
#
# =========================================================================== #
# require 'bioroebe/palindromes/palindrome_generator.rb'
# Bioroebe.palindrome(ARGV)
# =========================================================================== #
require 'bioroebe/base/commandline_application/commandline_application.rb'

module Bioroebe

class PalindromeGenerator < ::Bioroebe::CommandlineApplication # === Bioroebe::PalindromeGenerator

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      length_of_the_palindrome = nil,
      run_already              = true
    )
    reset
    set_length_of_the_palindrome(
      length_of_the_palindrome
    )
    if block_given?
      yielded = yield
      case yielded
      # ===================================================================== #
      # === :be_quiet
      # ===================================================================== #
      when :be_quiet
        set_be_quiet
      end
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @use_header_with_numbers
    #
    # If the following variable is set to true then a numbered header will
    # be shown above the sequence.
    # ======================================================================= #
    @use_header_with_numbers = true
    # ======================================================================= #
    # === @forward_strand
    # ======================================================================= #
    @forward_strand = ''.dup
    # ======================================================================= #
    # === @reverse_strand
    # ======================================================================= #
    @reverse_strand = ''.dup
  end

  # ========================================================================= #
  # === set_length_of_the_palindrome
  #
  # This setter-method determines the length of the palindrome that is to
  # be generated.
  # ========================================================================= #
  def set_length_of_the_palindrome(
      i = ''
    )
    i = i.first if i.is_a? Array
    i = i.to_i # Must be an Integer.
    if i.odd?
      if be_verbose?
        opnn { :short }
        e
        e
        erev 'The input '+sfancy(i.to_s)+rev+' is odd - we need an '\
             'even number for Palindromes.'
        erev 'Thus, we will deduct 1 from it and use that instead.'
      end
      i -= 1 # Odd numbers will become even here.
    end
    @length_of_the_palindrome = i
  end

  # ========================================================================= #
  # === length_of_the_palindrome?
  # ========================================================================= #
  def length_of_the_palindrome?
    @length_of_the_palindrome
  end

  # ========================================================================= #
  # === lpad?
  # ========================================================================= #
  def lpad?
    '  '
  end

  # ========================================================================= #
  # === report                                                   (report tag)
  #
  # You have to invoke the report method manually, thus, on your own.
  #
  # We will report first the forward strand and then the reverse strand.
  # ========================================================================= #
  def report
    e
    if @use_header_with_numbers
      print '       '
      n_times = @reverse_strand.size 
      n_times.times {|n| n += 1 # This begins at 0, hence we add +1.
        if n > 9
          n = n % 10
        end
        print steelblue(n)
      }
      e
    end
    report_forward_strand
    report_reverse_strand
    e
  end

  # ========================================================================= #
  # === report_forward_strand
  # ========================================================================= #
  def report_forward_strand(i = @forward_strand)
    e rev+lpad?+"5' - "+simp(i)+rev+" - 3'"
  end

  # ========================================================================= #
  # === report_reverse_strand
  # ========================================================================= #
  def report_reverse_strand(i = @reverse_strand)
    e rev+lpad?+"3' - "+simp(i)+rev+" - 5'"
  end

  # ========================================================================= #
  # === return_the_reverse_strand
  #
  # This will only return the reverse strand, but not assign to it.
  # ========================================================================= #
  def return_the_reverse_strand(
      forward_strand = @forward_strand
    )
    reverse_strand = ''.dup
    # ======================================================================= #
    # Ok, the forward strand was built - now we will add the reverse strand.
    # ======================================================================= #
    forward_strand.chars.each {|char|
      complement = find_complementary_nucleotide_to(char)
      reverse_strand << complement
    }
    return reverse_strand
  end

  # ========================================================================= #
  # === reverse_strand?
  # ========================================================================= #
  def reverse_strand?
    @reverse_strand
  end

  # ========================================================================= #
  # === forward_strand?
  # ========================================================================= #
  def forward_strand?
    @forward_strand
  end

  # ========================================================================= #
  # === result?
  # ========================================================================= #
  def result?
    @forward_strand+"\n"+@reverse_strand
  end

  # ========================================================================= #
  # === find_complementary_nucleotide_to
  #
  # Find the corresponding complementary nucleotide to the given token at
  # hand.
  # ========================================================================= #
  def find_complementary_nucleotide_to(i)
    return ::Bioroebe.complementary_nucleotide(i)
  end

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    first_half = ''.dup # This has the first half of the target.
    (length_of_the_palindrome? / 2).times {
      first_half << ::Bioroebe::DNA_NUCLEOTIDES.sample
    }
    forward_strand = first_half
    first_half.reverse.chars.each {|char|
      complement = find_complementary_nucleotide_to(char)
      forward_strand << complement
    }
    @forward_strand = forward_strand
    @reverse_strand = return_the_reverse_strand
  end

  # ========================================================================= #
  # === Bioroebe::PalindromeGenerator[]
  # ========================================================================= #
  def self.[](i = '')
    new(i)
  end

  # ========================================================================= #
  # === Bioroebe::PalindromeGenerator.return_as_string
  #
  # This method has been added particularly to embed the information
  # yielded by this class into a webpage (e. g. for sinatra). That
  # way we can make available the functionality for people who visit
  # a website making use of Bioroebe.
  #
  # Invocation example:
  #
  #   Bioroebe::PalindromeGenerator.return_as_string(6) # => "GTGCAC\nCACGTG\n"
  #
  # ========================================================================= #
  def self.return_as_string(i = ARGV)
    _ = ::Bioroebe::PalindromeGenerator.new(i) { :be_quiet }
    result = ''.dup
    result << _.forward_strand? << "\n" <<
              _.reverse_strand? << "\n"
    result
  end

end

# =========================================================================== #
# === Bioroebe.palindrome_generator
#
# Usage example:
#
#   puts Bioroebe.palindrome_generator(10)
#
# =========================================================================== #
def self.palindrome_generator(i = 6)
  ::Bioroebe::PalindromeGenerator.new(i).result?
end; self.instance_eval { alias palindrome palindrome_generator } # === Bioroebe.palindrome

end

if __FILE__ == $PROGRAM_NAME
  _ = Bioroebe::PalindromeGenerator.new(ARGV) { :be_quiet }
  _.report
end # palindromegenerator 16