#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === project.rb
#
# This file contains code that can be used for determining the
# base-directory of the BioRoebe project, as well as some constants
# for directories that directly depend on this, such as the
# constant called TOPLEVEL_METHODS_DIRECTORY.
#
# This file also contains a constant that points to the
# yaml-subdirectory of the bioroebe project, as well as
# for the Taxonomy submodule within module Bioroebe.
#
# It is recommended to name constants pointing to directories
# via a trailing "_DIRECTORY". In the past, this was often
# "_DIR" instead, but this is a bit less elegant to read
# in the code, in my opinion, so the longer variant should
# be used instead..
#
# Note that since past the year 2020 it is better to make use of
# methods rather than rely on hardcoded constants. While the old
# constants will not be deprecated (at the least not this time),
# it has been determined that using methods is, even though they
# are a tiny bit slower, more flexible, leading to code that
# can be maintained more easily as well in the long run.
# =========================================================================== #
# require 'bioroebe/project/project.rb'
# Bioroebe.project_yaml_directory?
# Bioroebe::CODON_TABLES_DIRECTORY
# =========================================================================== #
module Bioroebe

# =========================================================================== #
# === @project_base_directory
#
# This variable will keep track of the base directory for the project
# as-is.
# =========================================================================== #
@project_base_directory =
  File.absolute_path("#{__dir__}/..")+'/'

# =========================================================================== #
# === Bioroebe.project_base_dir?
#
# This method will return the value of @project_base_directory - that is
# where the files that belong to Bioroebe are assumed to reside.
# =========================================================================== #
def self.project_base_dir?
  @project_base_directory
end; self.instance_eval { alias project_dir?            project_base_dir? } # === Bioroebe.project_dir?
     self.instance_eval { alias project_base_directory? project_base_dir? } # === Bioroebe.project_base_directory?
     self.instance_eval { alias project_directory?      project_base_dir? } # === Bioroebe.project_directory?

# =========================================================================== #
# === Bioroebe.project_yaml_directory?
#
# This method will return the path to the yaml-directory of the bioroebe
# project. It will be a dynamic result rather than make use of the
# above constant.
# =========================================================================== #
def self.project_yaml_directory?(i = @project_base_directory)
  # ========================================================================= #
  # The following code allows us to default to the HOME directory of the
  # user.
  # ========================================================================= #
  use_this_as_the_project_base_directory = i
  unless File.directory? use_this_as_the_project_base_directory
    use_this_as_the_project_base_directory = File.expand_path('~')
  end
  return "#{use_this_as_the_project_base_directory}yaml/"
end; self.instance_eval { alias yaml_dir?       project_yaml_directory? } # === Bioroebe.yaml_dir?
     self.instance_eval { alias yaml_directory? project_yaml_directory? } # === Bioroebe.yaml_directory?

# =========================================================================== #
# === FILE_EXPAND_CD_ALIASES
# =========================================================================== #
FILE_EXPAND_CD_ALIASES =
  "#{@project_base_directory}hash_expand_cd_aliases.rb"

# =========================================================================== #
# === CONFIGURATION_DIRECTORY
# =========================================================================== #
CONFIGURATION_DIRECTORY = 
  "#{project_yaml_directory?}configuration/"

# =========================================================================== #
# === TOPLEVEL_METHODS_DIRECTORY
# =========================================================================== #
TOPLEVEL_METHODS_DIRECTORY =
  "#{@project_base_directory}toplevel_methods/"

# =========================================================================== #
# === CODON_TABLES_DIRECTORY
# =========================================================================== #
CODON_TABLES_DIRECTORY = 
  "#{@project_base_directory}codon_tables/"

# =========================================================================== #
# === CLEAVE_AND_DIGEST_DIRECTORY
# =========================================================================== #
CLEAVE_AND_DIGEST_DIRECTORY = 
  "#{@project_base_directory}cleave_and_digest/"

# =========================================================================== #
# === ELECTRON_MICROSCOPY_DIRECTORY
# =========================================================================== #
ELECTRON_MICROSCOPY_DIRECTORY = 
  "#{@project_base_directory}electron_microscopy/"

# =========================================================================== #
# === CODON_TABLES_DIRECTORY_FREQUENCY
# =========================================================================== #  
CODON_TABLES_DIRECTORY_FREQUENCY =
  "#{CODON_TABLES_DIRECTORY}frequencies/"

# =========================================================================== #
# === PDB_DIRECTORY
# =========================================================================== #
PDB_DIRECTORY =
  "#{@project_base_directory}pdb/"

# =========================================================================== #
# === PARSERS_DIRECTORY
# =========================================================================== #
PARSERS_DIRECTORY =
  "#{@project_base_directory}parsers/"

# =========================================================================== #
# === ENZYMES_DIRECTORY
# =========================================================================== #
ENZYMES_DIRECTORY =
  "#{@project_base_directory}enzymes/"

# =========================================================================== #
# === PALINDROMES_DIRECTORY
# =========================================================================== #
PALINDROMES_DIRECTORY =
  "#{@project_base_directory}palindromes/"

# =========================================================================== #
# === PATTERN_DIRECTORY
# =========================================================================== #
PATTERN_DIRECTORY =
  "#{@project_base_directory}pattern/"

# =========================================================================== #
# === NUCLEOTIDES_DIRECTORY
# =========================================================================== #
NUCLEOTIDES_DIRECTORY =
  "#{@project_base_directory}nucleotides/"

# =========================================================================== #
# === COUNT_DIRECTORY
# =========================================================================== #
COUNT_DIRECTORY =
  "#{@project_base_directory}count/"

# =========================================================================== #
# === AMINOACIDS_DIRECTORY
# =========================================================================== #
AMINOACIDS_DIRECTORY =
  "#{@project_base_directory}aminoacids/"

# =========================================================================== #
# === BLOSUM_DIRECTORY
# =========================================================================== #
BLOSUM_DIRECTORY =
  "#{project_yaml_directory?}blosum/"

# =========================================================================== #
# === CALCULATE_DIRECTORY
# =========================================================================== #
CALCULATE_DIRECTORY =
  "#{@project_base_directory}calculate/"

# =========================================================================== #
# === CODONS_DIRECTORY
# =========================================================================== #
CODONS_DIRECTORY =
  "#{@project_base_directory}codons/"

# =========================================================================== #
# === DOTPLOTS_DIRECTORY
# =========================================================================== #
DOTPLOTS_DIRECTORY =
  "#{@project_base_directory}dotplots/"

# =========================================================================== #
# === SEQUENCE_DIRECTORY
# =========================================================================== #
SEQUENCE_DIRECTORY =
  "#{@project_base_directory}sequence/"

# =========================================================================== #
# === PATHWAYS_DIRECTORY
#
# This constant will point to e. g.
# "/Programs/Ruby/2.6.4/lib/ruby/site_ruby/2.6.0/bioroebe/yaml/pathways/".
# =========================================================================== #
PATHWAYS_DIRECTORY =
  "#{project_yaml_directory?}pathways/"

# =========================================================================== #
# === BIOROEBE_YAML_AMINOACIDS_DIRECTORY
# =========================================================================== #
BIOROEBE_YAML_AMINOACIDS_DIRECTORY =
  "#{project_yaml_directory?}aminoacids/"

# =========================================================================== #
# === STRING_MATCHING_DIRECTORY
# =========================================================================== #
STRING_MATCHING_DIRECTORY =
  "#{@project_base_directory}string_matching/"

# =========================================================================== #
# === FASTA_AND_FASTQ_DIRECTORY
# =========================================================================== #
FASTA_AND_FASTQ_DIRECTORY =
  "#{@project_base_directory}fasta_and_fastq/"

# =========================================================================== #
# === Bioroebe::Taxonomy
#
# Add some code for the Taxonomy submodule next.
#
# The Taxonomy submodule will make use of the toplevel yaml/ subdirectory
# of the Bioroebe project in particular.
# =========================================================================== #
module Taxonomy

  # ========================================================================= #
  # === Bioroebe::Taxonomy::PROJECT_YAML_DIR
  # ========================================================================= #
  PROJECT_YAML_DIR = ::Bioroebe.project_base_directory?

  # ========================================================================= #
  # === Bioroebe::Taxonomy.project_base_dir?
  #
  # The Taxonomy project has been fully integrated into the Bioroebe
  # namespace in the year ~2015.
  #
  # This method will return a String such as:
  #
  #   "/Programs/Ruby/2.6.4/lib/ruby/site_ruby/2.6.0/bioroebe/taxonomy/"
  #
  # ========================================================================= #
  def self.project_base_dir?
    "#{::Bioroebe.project_base_directory?}taxonomy/"
  end

  # ========================================================================= #
  # === Bioroebe::Taxonomy.project_yaml_dir?
  # ========================================================================= #
  def self.project_yaml_dir?
    Taxonomy::PROJECT_YAML_DIR
  end

end; end

if __FILE__ == $PROGRAM_NAME
  puts Bioroebe.project_base_dir?
end