#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::ReverseComplement
#
# This class will turn input such as:
#
#   5'-AAAACCCGGT
#
# into output like:
#
#   3'-ACCGGGTTTT
#
# which is the reverse-complement, e. g. the sequence that is complementary
# to the above sequence. It can thus base-pair with the sequence given
# at 5'.
# =========================================================================== #
# require 'bioroebe/sequence/reverse_complement.rb'
# Bioroebe::ReverseComplement[ARGV]
# =========================================================================== #
require 'bioroebe/sequence/sequence.rb'

module Bioroebe

class ReverseComplement < Sequence # === Bioroebe::ReverseComplement

  alias e puts

  # ========================================================================= #
  # === DEFAULT_INPUT
  # ========================================================================= #
  DEFAULT_INPUT = 'AAAACCCGGT'

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      i           = DEFAULT_INPUT,
      run_already = true
    )
    reset
    set_sequence(i)
    # ======================================================================= #
    # === Handle blocks
    # ======================================================================= #
    if block_given?
      yielded = yield
      case yielded
      # ===================================================================== #
      # === :be_quiet
      # ===================================================================== #
      when :be_quiet
        @be_verbose = false
      end
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @be_verbose
    # ======================================================================= #
    @be_verbose = true
  end

  # ========================================================================= #
  # === complement?
  # ========================================================================= #
  def complement?
    @complement
  end; alias result?              complement? # === result?
       alias complement_sequence? complement? # === complement_sequence?

  # ========================================================================= #
  # === report                                                   (report tag)
  # ========================================================================= #
  def report
    e @complement if @be_verbose
  end

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    _ = sequence?
    # ======================================================================= #
    # (1) Obtain the reverse:
    # ======================================================================= #
    _ = _.reverse
    # ======================================================================= #
    # (2) Build the complement to that reverse:
    # ======================================================================= #
    @complement = ''.dup
    chars = _.chars
    # ======================================================================= #
    # The following code will build the reverse-complement.
    # ======================================================================= #
    chars.each {|character|
      case character
      when 'A'
        @complement << 'T'
      when 'T','U'
        @complement << 'A'
      when 'C'
        @complement << 'G'
      when 'G'
        @complement << 'C'
      end
    }
    report if @be_verbose
  end

  # ========================================================================= #
  # === Bioroebe::ReverseComplement[]
  # ========================================================================= #
  def self.[](i)
    sequence = new(i) { :be_quiet }.complement_sequence?
    return sequence
  end

end

# =========================================================================== #
# === Bioroebe.reverse_complement
#
# Invocation examples:
#
#   Bioroebe.reverse_complement 'AT'
#   Bioroebe.reverse_complement 'AAAACCCGGT' # => "ACCGGGTTTT"
#
# =========================================================================== #
def self.reverse_complement(i)
  ::Bioroebe::ReverseComplement.new(i) { :be_quiet }.result?
end

end

if __FILE__ == $PROGRAM_NAME
  Bioroebe::ReverseComplement[ARGV]
end # reversecomplement AAAACCCGGT