#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/atomic_composition.rb'
# Bioroebe.show_atomic_composition
# =========================================================================== #
module Bioroebe

  require 'bioroebe/constants/files_and_directories.rb'
  require 'bioroebe/colours/rev.rb'
  require 'bioroebe/toplevel_methods/aminoacids_and_proteins.rb'
  require 'bioroebe/toplevel_methods/e.rb'
  require 'bioroebe/toplevel_methods/is_on_roebe.rb'

  # ========================================================================= #
  # === Bioroebe.atomic_composition
  #
  # This method will return the composition of atoms in a given protein,
  # via Hash, such as:
  #
  #   {"C"=>11, "H"=>19, "N"=>5, "O"=>6, "S"=>0}
  #
  # The Hash keeps track of 11 C atoms, 19 H atoms, 5 N atoms,
  # 6 O atoms and 0 S atoms.
  #
  # This hash can then be formatted via the method:
  #
  #   Bioroebe.show_atomic_composition()
  #
  # Which can be found below.
  #
  # Presently this method works on aminoacids only, but in theory the
  # code could be extended to work with DNA nucleotides and RNA
  # nucleotides as well.
  #
  # Either way, the one letter abbreviation should be used as input
  # to this method.
  #
  # When we use aminoacids, we need to remember that a peptide bond
  # deducts 1x H₂O (water). This will have to be deducted from the
  # formula, but only if it is an internal aminoacid. In other words,
  # the only two aminoacids that will behave differently, are the
  # first one (since it will miss one -OH group) and the last
  # aminoacid (as this one will lack a -H molecule.
  #
  # Remember that the input sequence to this method should be the
  # one-letter code for the aminoacid sequence at hand.
  # ========================================================================= #
  def self.atomic_composition(
      of = 'GGGGA' # ← This should be the aminoacid sequence.
    )
    begin
      require 'chemistry_paradise/split_molecule_names.rb'
      require 'chemistry_paradise/toplevel_methods/remove_this_molecule_from.rb'
    rescue LoadError
      if is_on_roebe?
        puts 'Two files from the chemistry_paradise gem are not available.'
      end
    end
    # ======================================================================= #
    # Load up the molecular formula for each aminoacid next. This will
    # be used as our reference-point for calculating things such as the
    # composition, or weight.
    # ======================================================================= #
    dataset_molecular_formula_for_the_aminoacids = YAML.load_file(
      FILE_AMINOACIDS_MOLECULAR_FORMULA
    )
    if of.is_a?(Array)
      if of.empty?
        of = 'GGGGA' # In this case reinstate the default.
      else
        if of.first.is_a?(String) and of.first.size > 1
          of = of.first.split(//) # Split it on a per-character basis here.
        end
      end
    end
    if of.is_a? String
      of = of.split(//)
    end
    unless of.is_a? Array
      of = [of]
    end
    hash_keeping_track_of_the_atomic_composition = {}
    # ======================================================================= #
    # Build up the default values, for the atoms C, H, N, O and S.
    # ======================================================================= #
    hash_keeping_track_of_the_atomic_composition['C'] = 0
    hash_keeping_track_of_the_atomic_composition['H'] = 0
    hash_keeping_track_of_the_atomic_composition['N'] = 0
    hash_keeping_track_of_the_atomic_composition['O'] = 0
    hash_keeping_track_of_the_atomic_composition['S'] = 0
    # ======================================================================= #
    # Next obtain the formula from the ChemistryParadise project. We
    # do so by iterating over the given input, and we assume that
    # this input is always an Array.
    # ======================================================================= #
    of.map.with_index {|this_amino_acid, position_of_that_aminoacid|
      # ===================================================================== #
      # Next, we have to obtain the formula for this amino acid.
      # ===================================================================== #
      this_amino_acid = AMINO_ACIDS_ENGLISH[this_amino_acid]
      formula_for_this_amino_acid = dataset_molecular_formula_for_the_aminoacids[this_amino_acid]
      # ===================================================================== #
      # The next case-menu will handle the position of the aminoacid at hand.
      # We will skip doing so if there is only one aminoacid though.
      # ===================================================================== #
      if of.first.to_s.size > 1
        case position_of_that_aminoacid # case tag
        when 0 # This is the first aminoacid. It loses only one 'OH' group.
          formula_for_this_amino_acid = 
            ::ChemistryParadise.remove_this_molecule_from('OH', formula_for_this_amino_acid)
        when (of.size - 1) # This is the last entry. It loses only one 'H' group.
          formula_for_this_amino_acid = 
            ::ChemistryParadise.remove_this_molecule_from('H', formula_for_this_amino_acid)
        else
          # ================================================================= #
          # Else it will lose a full H₂O group.
          # ================================================================= #
          formula_for_this_amino_acid = 
            ::ChemistryParadise.remove_this_molecule_from('H2O', formula_for_this_amino_acid)
        end
      end
      array_chemical_formula = ::ChemistryParadise.split_this_molecular_formula_into_a_hash(
        formula_for_this_amino_acid
      )
      array_chemical_formula.each {|molecule_and_number| # e. g. 'H13'
        if molecule_and_number =~ /\d+/ # If it has at the least one number.
          molecule_and_number =~ /([A-Z]+)(\d{1,2})/ # See: https://rubular.com/r/nCojEDcY6g
          molecule = $1.to_s.dup
          n_times  = $2.to_s.dup.to_i
          hash_keeping_track_of_the_atomic_composition[molecule] += n_times
        else # else it must be 1, since there is no other number, such as 'N'.
          hash_keeping_track_of_the_atomic_composition[molecule_and_number] += 1
        end
      }
    }
    return hash_keeping_track_of_the_atomic_composition
  end

  # ========================================================================= #
  # === Bioroebe.show_atomic_composition
  #
  # This method will output the above.
  #
  # Display style is similar to ExPASY output.
  #
  #   Atomic composition:
  #
  #   Carbon      C	        11
  #   Hydrogen    H	        19
  #   Nitrogen    N	         5
  #   Oxygen      O	         6
  #   Sulfur      S	         0
  #
  # ========================================================================= #
  def self.show_atomic_composition(
      of = 'GGGGA'
    )
    if of.is_a? Array
      of = of.join(' ').strip
    end
    padding_to_use = '  '
    hash = ::Bioroebe.atomic_composition(of)
    _ = ''.dup
    _ << "#{rev}Atomic composition of #{steelblue(of)}#{rev}:\n\n"
    n_atoms_in_total = hash.values.sum
    array = %w( C H N O S )
    array.each {|this_atom|
      case this_atom
      when 'C'
        _ << "#{padding_to_use}#{rev}Carbon    C  #{hash[this_atom].to_s.rjust(4)}\n"
      when 'H'
        _ << "#{padding_to_use}#{rev}Hydrogen  H  #{hash[this_atom].to_s.rjust(4)}\n"
      when 'N'
        _ << "#{padding_to_use}#{rev}Nitrogen  N  #{hash[this_atom].to_s.rjust(4)}\n"
      when 'O'
        _ << "#{padding_to_use}#{rev}Oxygen    O  #{hash[this_atom].to_s.rjust(4)}\n"
      when 'S'
        _ << "#{padding_to_use}#{rev}Sulfur    S  #{hash[this_atom].to_s.rjust(4)}\n"
      end
    }
    _ << "  "+("-" * 18)
    _ << "\n"
    _ << "  Total:          #{n_atoms_in_total}\n\n"
    puts _
  end

end

if __FILE__ == $PROGRAM_NAME
  if ARGV.empty?
    pp Bioroebe.atomic_composition(ARGV)
    Bioroebe.show_atomic_composition
  else
    Bioroebe.show_atomic_composition(ARGV)
  end
end # atomiccomposition GGGGA