#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/base_composition.rb'
# Bioroebe.report_base_composition
# =========================================================================== #
module Bioroebe

  require 'bioroebe/colours/colours.rb'
  require 'bioroebe/colours/rev.rb'
  require 'bioroebe/toplevel_methods/e.rb'

  # ========================================================================= #
  # === Bioroebe.base_composition
  #
  # This method can be used to query the composition of a given DNA
  # sequence, that is, in percentage, the values for A, T, C and
  # G.
  #
  # This method will then return a Hash, consisting of the percentage
  # values of A, T, C and G in the given DNA sequence at hand.
  #
  # Note that the input to this method has to include a '%' character,
  # at the least up until March 2020. Past March 2020 this requirement 
  # was dropped, but I still think it is visually more elegant to
  # include a '%' character.
  # ========================================================================= #
  def self.base_composition(
      i = '52%GC'
    )
    if i.is_a? Array
      if i.empty?
        i = '52%GC' # Default value.
      else
        i = i.join(' ').strip
      end
    end
    # ======================================================================= #
    # We must use a Hash for this.
    # ======================================================================= #
    hash = {
      'A' => 0,
      'T' => 0,
      'C' => 0,
      'G' => 0,
    }
    if i.include? '%'
      splitted = i.split('%').map(&:strip)
      frequency = splitted.first.to_i
      opposite_frequency = 100 - frequency
      characters = splitted.last.split(//)
      characters.each {|this_nucleotide|
        hash[this_nucleotide] = frequency / 2
      }
      # ===================================================================== #
      # Next calculate the missing nucleotides:
      # ===================================================================== #
      missing_nucleotides = hash.select {|key, value|
        value == 0
      }
      missing_nucleotides.each_pair {|this_nucleotide, value|
        hash[this_nucleotide] = opposite_frequency / 2
      }
    else
      frequency = Hash.new(0)
      chars = i.chars
      chars.each { |entry| frequency[entry] += 1 }
      sum = frequency.values.sum
      frequency.each_pair {|this_nucleotide, value|
        hash[this_nucleotide] = ((value * 100.0) / sum).round(2)
      }
    end
    return hash
  end; self.instance_eval { alias frequency_of base_composition } # === Bioroebe.frequency_of

  # ========================================================================= #
  # === Bioroebe.report_base_composition
  #
  # This is the convenience-method for reporting the base composition
  # at hand.
  # ========================================================================= #
  def self.report_base_composition(
      i = '52%GC'
    )
    if i.is_a? Array
      if i.empty?
        i = '52%GC' # Default value.
      else
        i = i.join(' ').strip
      end
    end
    hash = ::Bioroebe.base_composition(i)
    a = hash['A']
    t = hash['T']
    c = hash['C']
    g = hash['G']
    erev 'The base composition frequencies of this sequence '\
         '(length: '+i.size.to_s+') is as follows:'
    e
    e '  '\
      ' A: '+steelblue(a.to_s+'%')+rev+
      ' T: '+steelblue(t.to_s+'%')+rev+
      ' C: '+steelblue(c.to_s+'%')+rev+
      ' G: '+steelblue(g.to_s+'%')+rev
    e
  end; self.instance_eval { alias show_base_composition report_base_composition } # === show_base_composition

end

if __FILE__ == $PROGRAM_NAME
  pp Bioroebe.base_composition(ARGV)
  Bioroebe.report_base_composition(ARGV)
end # basecomposition ATGC
    # basecomposition ATTGCCCCCCCCAACCCCACA