#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/download_and_fetch_data.rb'
# =========================================================================== #
module Bioroebe

  require 'bioroebe/toplevel_methods/e.rb'
  require 'bioroebe/toplevel_methods/esystem.rb'
  require 'bioroebe/toplevel_methods/fasta_and_fastq.rb'
  require 'bioroebe/toplevel_methods/log_directory.rb'
  require 'bioroebe/toplevel_methods/file_and_directory_related_actions.rb'
  require 'bioroebe/colours/colours.rb'

  # ========================================================================= #
  # === Bioroebe.fetch_data_from_uniprot
  #
  # The first argument to this method should be the name of the protein
  # at hand. Alternatively it can also be just the full URL.
  #
  # The second argument specifies whether we will automatically try
  # to rename this FASTA file. By default this is enabled, largely
  # because it makes it a bit easier to know what a particular
  # fasta sequence contains.
  # ========================================================================= #
  def self.fetch_data_from_uniprot(
      i = 'B5ZC00',
      do_perform_rename_action = true
    )
    require 'open-uri'
    # ======================================================================= #
    # Support Array as input:
    # ======================================================================= #
    if i.is_a? Array
      i.each {|entry|
        fetch_data_from_uniprot(entry, do_perform_rename_action)
      }
    else
      if i.frozen?
        i = i.dup
      end
      i.strip!
      if i.start_with? 'http'
        remote_url = i
      else
        remote_url = "https://www.uniprot.org/uniprot/#{i}".dup
        remote_url << '.fasta' unless remote_url.end_with? '.fasta'
      end
      # ======================================================================= #
      # Store this protein sequence at an appropriate location:
      # ======================================================================= #
      e
      erev "Trying to obtain remote data from "\
           "#{sfancy(remote_url)} #{rev}next:"
      e
      store_here = LOCAL_DIRECTORY_FOR_UNIPROT+
                   File.basename(remote_url)
      # ======================================================================= #
      # Next create the directory unless it already exists.
      # ======================================================================= #
      unless File.directory? LOCAL_DIRECTORY_FOR_UNIPROT
        mkdir LOCAL_DIRECTORY_FOR_UNIPROT
      end
      dataset = URI.open(remote_url).read
      e "The remote dataset will be stored here: "\
        "#{sfile(store_here)}"
      e
      write_what_into(dataset, store_here)
      # ======================================================================= #
      # Next comes a rename action - this will make the .fasta file more
      # verbose, but I think this may still be helpful, in particular if you
      # have a lot of .fasta files. There is a switch that controls this
      # setting though.
      # ======================================================================= #
      if do_perform_rename_action
        # ===================================================================== #
        # The initially assumed filename may be like this:
        #
        #   /root/Bioroebe/A2Z669_CSPLT_ORYSI_CASP-like_protein_5A2_OS=Oryza_sativa_subsp_indica_OX=39946_GN=OsI_33147_PE=3_SV=1.fasta
        #
        # ===================================================================== #
        assumed_filename = ::Bioroebe.return_new_filename_based_on_fasta_identifier(store_here)
        new_filename = LOCAL_DIRECTORY_FOR_UNIPROT+
                       File.basename(assumed_filename)
        erev "Next renaming `#{sfile(store_here)}#{rev}` to"
        erev "  `#{sfile(new_filename)}`."
        ::Bioroebe.mv(store_here, new_filename)
        return new_filename
      end
      return store_here
    end
  end

  # ========================================================================= #
  # === Bioroebe.wget_download
  #
  # This variant will always use "wget" - hence the name.
  # ========================================================================= #
  def self.wget_download(target)
    esystem "wget #{target}"
  end

  # ========================================================================= #
  # === Bioroebe.download
  # ========================================================================= #
  def self.download(
      from_these_URLs
    )
    require 'open-uri'
    array_these_urls = [from_these_URLs].flatten.compact
    array_these_urls.each {|remote_url|
      # ===================================================================== #
      # First, we must determine the remote file listing here.
      # Due to convenience we will simply use curl here.
      # ===================================================================== #
      cmd = "curl -s \"#{remote_url}\" --list-only"
      # e cmd
      remote_files = `#{cmd}`.split("\n")
      remote_files.each {|this_remote_file|
        target = remote_url+this_remote_file
        e "Downloading `#{this_remote_file}` next. '"\
          "(Full target: '#{target})"
        wget_download(target)
      }
    }
  end

end

if __FILE__ == $PROGRAM_NAME
  Bioroebe.download(ARGV)
  Bioroebe.fetch_data_from_uniprot(ARGV)
end # bioroebedownload ftp://ftp.ensembl.org/pub/release-92/gtf/mus_musculus/
    # fetchdatafromuniprot
    # fetchdatafromuniprot B5ZC00
    # fetchdatafromuniprot A2Z669
    # fetchdatafromuniprot P07204_TRBM_HUMAN
    # fetchdatafromuniprot P20840_SAG1_YEAST 