#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/file_and_directory_related_actions.rb'
# =========================================================================== #
module Bioroebe

  require 'fileutils'
  require 'bioroebe/toplevel_methods/e.rb'
  require 'bioroebe/toplevel_methods/log_directory.rb'
  require 'bioroebe/colours/rev.rb'
  require 'bioroebe/colours/sfile.rb'
  require 'bioroebe/colours/sdir.rb'

  # ========================================================================= #
  # === Bioroebe.pwd                                                (pwd tag)
  #
  # Feedback the current working directory (it will be returned, as a
  # String).
  # ========================================================================= #
  def self.pwd
    "#{Dir.pwd}/".squeeze('/')
  end; self.instance_eval { alias return_pwd pwd } # === Bioroebe.return_pwd

  # ========================================================================= #
  # === pwd
  # ========================================================================= #
  def pwd
    ::Bioroebe.pwd
  end

  # ========================================================================= #
  # === Bioroebe.no_file_exists_at
  # ========================================================================= #
  def self.no_file_exists_at(i)
    erev "No file exists at `#{sfile(i)}#{rev}`."
  end

  # ========================================================================= #
  # === Bioroebe.mv
  #
  # This method can be used to rename or move a (local) file.
  # ========================================================================= #
  def self.mv(old, new)
    FileUtils.mv(old, new)
  end; self.instance_eval { alias rename mv } # === Bioroebe.rename

  # ========================================================================= #
  # === Bioroebe.create_file
  #
  # This method can be used to create a file.
  # ========================================================================= #
  def self.create_file(i)
    FileUtils.touch(i) unless File.file?(i)
  end

  # ========================================================================= #
  # === Bioroebe.append_what_into
  #
  # This method can be used to append content onto a file.
  # ========================================================================= #
  def self.append_what_into(
      what = 'Hello world!',
      into = 'test.md'
    )
    unless File.exist? into
      base_dir = File.dirname(into)
      unless File.directory? base_dir
        e rev+
        'No directory exists at '+sdir(base_dir)+
        rev+'. Thus creating it now.'
        create_directory(base_dir)
      end
      e rev+
        'No file exists at '+sfile(into)+rev+
        '. Thus creating it now.'
      create_file(into)
    end
    File.open(into, 'a') { |file|
      file << what
    }
  end

  # ========================================================================= #
  # === Bioroebe.write_what_into
  #
  # This method will simply write the content of the variable `what`
  # into the file called `into`.
  # ========================================================================= #
  def self.write_what_into(
      what = 'Hello world!',
      into = 'test.md'
    )
    base_directory = File.dirname(into)
    if File.directory? base_directory
      File.open(into, 'w') { |file|
        file << what
      }
    else
      e "No directory exists at #{base_directory}. Thus, "\
        "no file could be saved."
    end
  end; self.instance_eval { alias save_file       write_what_into } # === Bioroebe.save_file
       self.instance_eval { alias save_what_where write_what_into } # === Bioroebe.save_what_where

  # ========================================================================= #
  # === Bioroebe.remove_file
  #
  # Use this method to remove a file.
  # ========================================================================= #
  def self.remove_file(
      i, be_verbose = false
    )
    if File.exist? i
      File.delete(i)
      if be_verbose
        this_file = "#{return_pwd}#{i}"
        e 'File '+sfile(this_file)+' exists, '\
          'we will thus remove it now.'
      end
    end
  end; self.instance_eval { alias remove remove_file } # === Bioroebe.remove

  # ========================================================================= #
  # === Bioroebe.change_directory
  #
  # This method allows us to change the directory.
  #
  # Bioroebe.cd() is an alias to the method here.
  # ========================================================================= #
  def self.change_directory(
      i          = '$HOME',
      be_verbose = false
    )
    case be_verbose
    when :do_report_current_directory,
         :be_verbose
      be_verbose = true
    end
    case i # Do some sanitizing here. (case tag)
    # ======================================================================= #
    # === :home_directory
    # ======================================================================= #
    when :home_directory,
         :default,
         nil # ← Nil is also assumed to refer to this :default value.
      # ===================================================================== #
      # In this case we will try to cd into the base-directory of the
      # Bioroebe shell.
      # ===================================================================== #
      i = log_dir?
    # ======================================================================= #
    # === :download_dir
    # ======================================================================= #
    when :download_dir,':download_dir',
         :download_directory,':download_directory'
      i = download_dir?
    # ======================================================================= #
    # Bioroebe.save_dir? is defined in bioroebe/toplevel_methods/store_here.rb.
    # ======================================================================= #
    when 'base',
         'logdir'
      # ===================================================================== #
      # Enter the main log dir, unless a file exists with the same name.
      # ===================================================================== #
      i = save_dir? unless File.exist? i
    end
    i << '/' unless i.end_with? '/'
    if File.directory? i
      e sdir(i) if be_verbose # Also colourize the directory and output it.
      Dir.chdir(i)
    else
      if be_verbose
        erev "No directory called `#{sdir(i)}#{rev}` exists,"
        erev 'thus we can not cd to this target.'
      end
    end
  end; self.instance_eval { alias cd    change_directory } # === Bioroebe.cd
       self.instance_eval { alias chdir change_directory } # === Bioroebe.chdir

  # ========================================================================= #
  # === Bioroebe.pdb_directory?
  #
  # This refers to the local pdb/ directory.
  # ========================================================================= #
  def self.pdb_directory?
    return "#{log_dir?}pdb/"
  end

  # ========================================================================= #
  # === Bioroebe.ensure_that_the_base_directories_exist
  #
  # This method will ensure that the base directory for the
  # Bioroebe project exist.
  # ========================================================================= #
  def self.ensure_that_the_base_directories_exist
    # ======================================================================= #
    # We also need to create the temp directory, as well as having to
    # notify the user that this will be done. The taxonomy subdirectory
    # will also be created.
    # ======================================================================= #
    use_this_log_dir = log_dir?
    unless File.exist? use_this_log_dir
      erev "The base directory at `#{sdir(use_this_log_dir)}#{rev}` does not exist."
      erev 'It will thus be created next.'
      mkdir use_this_log_dir
    end
    # ======================================================================= #
    # === Ensure that the Downloads/ directory exists
    # ======================================================================= #
    _ = "#{use_this_log_dir}Downloads/"
    unless File.exist? _
      erev "The directory at `#{sdir(_)}#{rev}` does not exist."
      erev 'It will thus be created next.'
      mkdir _
    end
    # ======================================================================= #
    # === Ensure that the pdb/ directory exists
    # ======================================================================= #
    _ = "#{use_this_log_dir}pdb/"
    unless File.exist? _
      erev "The directory at `#{sdir(_)}#{rev}` does not exist."
      erev 'It will thus be created next.'
      mkdir _
    end
    autogenerated_sql_files_dir =
      AUTOGENERATED_SQL_FILES_DIR
    unless Dir.exist? autogenerated_sql_files_dir
      erev 'The directory at `'+sdir(autogenerated_sql_files_dir)+
           rev+'` does not exist.'
      erev 'It will thus be created next.'
      mkdir(autogenerated_sql_files_dir)
    end
  end

  # ========================================================================= #
  # === Bioroebe.mkdir                                            (mkdir tag)
  # ========================================================================= #
  def self.mkdir(i)
    FileUtils.mkdir_p(i)
  end; self.instance_eval { alias create_directory mkdir } # === Bioroebe.create_directory
       self.instance_eval { alias mkdir_p          mkdir } # === Bioroebe.mkdir_p

  # ========================================================================= #
  # === mkdir
  # ========================================================================= #
  def mkdir(i)
    ::Bioroebe.mkdir(i)
  end

end

if __FILE__ == $PROGRAM_NAME
  Bioroebe.ensure_that_the_base_directories_exist
  Bioroebe.write_what_into(ARGV[0], ARGV[1])
end