#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/misc.rb'
# Bioroebe.reverse_sequence(ARGV)
# Bioroebe.bisulfite_treatment(ARGV)
# Bioroebe.return_DNA_composition_hash
# =========================================================================== #
module Bioroebe

  require 'bioroebe/encoding/encoding.rb'
  require 'bioroebe/toplevel_methods/e.rb'
  require 'bioroebe/toplevel_methods/aminoacids_and_proteins.rb'

  # ========================================================================= #
  # === Bioroebe.return_ubiquitin_sequence
  #
  # This method will simply return the Ubiquitin sequence. We may also
  # colourize some aminoacids, if we have enabled this.
  # ========================================================================= #
  def self.return_ubiquitin_sequence(
      i = 'MQIFVKTLTGKTITLEVEPSDTIENVKAKIQDKEGIPPDQQRLIFAGKQLEDGRTLSDYNIQKESTLHLVLRLRGG'
    )
    colourize_aa(i) # Consider colourizing it.
  end; self.instance_eval { alias ubiquitin return_ubiquitin_sequence } # === Bioroebe.ubiquitin

  # ========================================================================= #
  # === Bioroebe.reverse_sequence
  # ========================================================================= #
  def self.reverse_sequence(i)
    i = i.first if i.is_a? Array
    i.reverse
  end

  # ========================================================================= #
  # === Bioroebe.bisulfite_treatment
  #
  # Simply convert all C into U.
  # ========================================================================= #
  def self.bisulfite_treatment(i)
    if i.is_a? Array
      i = i.join('').strip
    end
    i.tr('C','U')
  end; self.instance_eval { alias bisulfie bisulfite_treatment } # === Bioroebe.bisulfite

  # ========================================================================= #
  # === Bioroebe.batch_create_windows_executables
  #
  # This method is only useful for windows. We will use "ocra" to create
  # various .exe files that have the desired widgt-functionality.
  #
  # Note that the functionality depends on the roebe-gem.
  # ========================================================================= #
  def self.batch_create_windows_executables
    begin
      require 'roebe/custom_methods/custom_methods.rb'
    rescue LoadError; end
    array_these_files =  %w(
      /home/x/programming/ruby/src/bioroebe/lib/bioroebe/gui/libui/hamming_distance/hamming_distance.rb
    )
    array_these_files.each {|this_file|
      Roebe.ocra_build(
        this_file,
        File.basename(this_file).delete_suffix('.rb')
      )
    }
  end

  # ========================================================================= #
  # === Bioroebe.create_the_pdf_tutorial
  #
  # This method can be used to quickly turn the README.md file into
  # a .pdf file, for whatever the reason the user may want this.
  # ========================================================================= #
  def self.create_the_pdf_tutorial(
      read_from_this_file = '/home/x/programming/ruby/src/bioroebe/README.md',
      store_where         = '/Depot/j/example.pdf'
    )

    require 'prawn'

    Prawn::Fonts::AFM.hide_m17n_warning = true # Hide a useless warning.

    pdf = Prawn::Document.new(
            page_size: 'A4',
            page_layout: :landscape
          )
    pdf.text "The Bioroebe Project", size: 80
    pdf.start_new_page
    pdf.bounding_box [50, 600], width: 200 do
      pdf.fill_color '000000'
      pdf.text "making bioinformatics great again:", size: 15
    end
    pdf.start_new_page
    dataset = File.read(read_from_this_file, encoding: UTF_ENCODING)
    dataset = dataset.encode("Windows-1252", invalid: :replace, undef: :replace)

    pdf.text(dataset)
    e 'Storing at this location: '+store_where
    pdf.render_file store_where
  end; self.instance_eval { alias pdf create_the_pdf_tutorial } # === Bioroebe.pdf

  # ========================================================================= #
  # === Bioroebe.return_DNA_composition_hash
  # ========================================================================= #
  def self.return_DNA_composition_hash(
      of_this_sequence
    )
    of_this_sequence = of_this_sequence.first if of_this_sequence.is_a? Array
    hash = {}
    array1 = %w( A T C G )
    array2 = Array.new(array1.size, 0)
    array = array1.zip(array2) # Zip it up with 0-values here.
    hash = Hash[array]
    of_this_sequence.chars.each {|this_char|
      hash[this_char] += 1
    }
    sorted_hash = Hash[*hash.sort_by {|key, value|
      key
    }.flatten]
    # ======================================================================= #
    # The hash may then look like this:
    #
    #   {"A"=>7, "T"=>6, "C"=>7, "G"=>6}
    #
    # ======================================================================= #
    return sorted_hash
  end

end

if __FILE__ == $PROGRAM_NAME
  puts Bioroebe.reverse_sequence(ARGV)
  puts Bioroebe.bisulfite_treatment('CCCGCAATGCATACCTCGCCG')
  puts Bioroebe.return_ubiquitin_sequence
  pp Bioroebe.return_DNA_composition_hash(ARGV)
end # rb misc ATGC # => CGTA
    # rb misc.rb GATCTGCAGTACGTACGTACTGACAC
    # returnubiquitinsequence