#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/number_of_clones.rb'
# =========================================================================== #
module Bioroebe

  # ========================================================================= #
  # === Bioroebe.number_of_clones
  #
  # This method will "calculate" the number of clones required for
  # an experiment that attempts to create a genomic library.
  #
  # The general formula is:
  #
  #   N = ln (1 - P ) / ln (1 - a/b)
  #
  # Where:
  #
  #   N is the number of clones required.
  #   P is the desired probability.
  #   a is the average size of the DNA fragments that are cloned.
  #   b is the size of the genome at hand (in n bp).
  #
  # ========================================================================= #
  def self.number_of_clones(
      average_size_of_the_dna_fragments = 10_000,
      size_of_the_genome                = 4_693_221, # ← E. coli K12.
      desired_probability               = 0.99       # ← Or 0.95 (this is P)
    )
    n_clones = Math.log(1 - desired_probability) /
              (Math.log(1 - (average_size_of_the_dna_fragments.to_f / size_of_the_genome.to_f)))
    n_clones.round(0)
  end

end

if __FILE__ == $PROGRAM_NAME
  require 'colours/autoinclude'
  genome_size = 4.0 * (10 ** 3)
  probability = 0.95
  insert_size = 20 # 20kb, lambda vector
  e 'For a probability of '+probability.to_s+', we will obtain these results:'
  e
  e '  For E. coli (insert size '+insert_size.to_s+' kb): '+
    Bioroebe.number_of_clones(insert_size, genome_size, probability).to_s+
    ' clones are required' # 598
  insert_size = 45 # 45kb, cosmid vector
  e '  For E. coli (insert size '+insert_size.to_s+' kb): '+
    Bioroebe.number_of_clones(insert_size, genome_size, probability).to_s+
    ' clones are required' # 598
  genome_size = 1.4 * (10 ** 4)
  insert_size = 20 # 20kb, lambda vector
  e '  For S. cerevisiae (insert size '+insert_size.to_s+' kb): '+
    Bioroebe.number_of_clones(insert_size, genome_size, probability).to_s+
    ' clones are required' # 598
  insert_size = 45 # 45kb, cosmid vector
  e '  For S. cerevisiae (insert size '+insert_size.to_s+' kb): '+
    Bioroebe.number_of_clones(insert_size, genome_size, probability).to_s+
    ' clones are required' # 598
  e
end # numberofclones