#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/toplevel_methods/phred_error_probability.rb'
# Bioroebe.phred_error_probability(ARGV)
# =========================================================================== #
module Bioroebe

  # ========================================================================= #
  # === Bioroebe.phred_error_probability
  #
  # Datasets obtained from next-generation reads specify a Phred score
  # for each base. This is also known as the "Q score" (Quality score).
  #
  # The Q score is an integer value, and is typically within the range 2
  # to 40. Q indicates the probability that the base call is incorrect (P_e).
  #
  # For example, Q=2 means that the error probability is 63%, so the machine
  # is reporting that the base is more likely to be wrong than right, while
  # Q=20 corresponds to an error probability of only 1%.
  #
  # The second argument to this method can be used to format to a specific
  # number. So if we pass in 6 as the second argument then the result will
  # be rounded to 6 slots after the first '.'. See the usage examples
  # below.
  #
  # Usage examples:
  #
  #   Bioroebe.phred_error_probability '48'
  #   Bioroebe.phred_error_probability '48',6 # => "0.000016"
  #
  # ========================================================================= #
  def self.phred_error_probability(
      quality_score                = 50,
      optional_use_this_formatting = nil
    )
    if quality_score.is_a? Array
      quality_score = quality_score.first
    end
    quality_score = quality_score.to_f # Need a Float past this point.
    result = 10 ** (- quality_score / 10.0)
    # ======================================================================= #
    # Next, do some rounding if the user requested this through the
    # second argument to this method.
    # ======================================================================= #
    if optional_use_this_formatting
      if optional_use_this_formatting.is_a? Numeric
        use_this_formatting_rule = "%.#{optional_use_this_formatting}f"
        result = use_this_formatting_rule % result
      else
        result = optional_use_this_formatting.to_s % result
      end
    end
    return result
  end

end

if __FILE__ == $PROGRAM_NAME
  if ARGV.empty?
  (1 .. 40).each {|entry|
    puts "#{entry.to_s.ljust(3)} #{Bioroebe.phred_error_probability(entry)}"
  }
  else
    puts Bioroebe.phred_error_probability(ARGV)
  end
end # phrederrorprobability 48