#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# The file here will contain code that is related to "find"-ing something,
# typically within another string or object or dataset.
# =========================================================================== #
# require 'bioroebe/toplevel_methods/searching_and_finding.rb'
# Bioroebe.find_substring 'GATATATGCATATACTT', 'ATAT' # => [2, 4, 10]
# =========================================================================== #
module Bioroebe

  # ========================================================================= #
  # === Bioroebe.find_substring
  #
  # This method can be used to find a substring within a larger String.
  #
  # For example, in the below default values, the substring "ATAT"
  # would exist at the positions 2, 4 and 10, if compared to the
  # larger parent string "GATATATGCATATACTT".
  #
  # The following display may help you see this more easily, in
  # regards to the substring matches:
  #
  #   GATATATGCATATACTT
  #    ATATAT  ATAT
  #
  # If you look closely, you will be able to see that "ATAT" can
  # be found three times in the string above.
  #
  # Indices in this context start at position 1, not 0. This is
  # mostly done to refer to nucleotides or aminoacids, which
  # also typically start at the first letter. Position 0 makes
  # no sense for a nucleotide - what would "nucleotide 0" even
  # refer to?
  #
  # The first argument to this method may also be the path to a
  # locally existing file, such as "/rosalind_subs.txt". In
  # fact this method has been largely motivated by Rosalind
  # tasks.
  #
  # The method will return an Array with the positions of all
  # substrings that are found in the full_string variable. See
  # the usage example below for how this may be.
  #
  # Usage example:
  #
  #    Bioroebe.find_substring 'GATATATGCATATACTT', 'ATAT' # => [2, 4, 10]
  #
  # ========================================================================= #
  def self.find_substring(
      full_string    = 'GATATATGCATATACTT', # ← The full String comes here.
      this_substring = :default             # ← The substring we are searching for comes here.
    )
    if full_string.is_a? Array
      # ===================================================================== #
      # Presently this method will only work on the first member of an Array.
      # ===================================================================== #
      full_string = full_string.first
    end
    if full_string and File.file?(full_string) and
       this_substring == :default
      # ===================================================================== #
      # In this case it is ok to read from that file.
      # ===================================================================== #
      _ = File.read(full_string)
      splitted = _.split("\n")
      full_string    = splitted.first
      this_substring = splitted.last
    end
    case this_substring
    # ======================================================================= #
    # Use a default value in this case. In reality users should supply
    # their own substring when they use this method here.
    # ======================================================================= #
    when :default,
         nil
      this_substring = 'ATAT'
    else
      if this_substring.empty?
        this_substring = 'ATAT'
      end
    end
    if full_string.nil? or full_string.empty?
      full_string = 'GATATATGCATATACTT' # ← Use the default in this case.
    end
    result = Array.new.tap { |indexes|
      final_index_position = full_string.size - this_substring.size
      i = 0
      while (i < final_index_position)
        index = full_string.to_s.index(this_substring.to_s, i)
        break if index.nil?
        i = index + 1
        indexes << i
      end
    }
    result = nil if result.empty? # ← We will try this here; could also return an empty Array, though.
    result # Return our findings here.
  end; self.instance_eval { alias find_substring_indices find_substring } # === Bioroebe.find_substring_indices

end

if __FILE__ == $PROGRAM_NAME
  if File.exist?(ARGV.first)
    # ======================================================================= #
    # Handle locally existing files first.
    # ======================================================================= #
    puts Bioroebe.find_substring(ARGV.first).join(' ').strip
  else
    puts Bioroebe.find_substring(ARGV[0], ARGV[1]).join(' ').strip
  end
end # findsubstring TATATA AT
    # findsubstring TATATATA AT
    # findsubstring GATATATGCATATACTT ATAT
    # findsubstring GATATATGCATATACTT ATAT
    # findsubstring /rosalind_subs.txt