#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# Taxonomy-related code can be stored in this toplevel .rb file.
# =========================================================================== #
# require 'bioroebe/toplevel_methods/taxonomy.rb'
# =========================================================================== #
module Bioroebe

module Taxonomy # === Bioroebe::Taxonomy

  require 'bioroebe/constants/newline.rb'
  require 'bioroebe/constants/misc.rb'
  require 'bioroebe/colours/colours.rb'
  require 'bioroebe/colours/rev.rb'
  require 'bioroebe/toplevel_methods/download_and_fetch_data.rb'
  require 'bioroebe/toplevel_methods/extract.rb'
  require 'bioroebe/toplevel_methods/file_and_directory_related_actions.rb'
  require 'bioroebe/toplevel_methods/log_directory.rb'
  require 'bioroebe/toplevel_methods/time_and_date.rb'

  # ========================================================================= #
  # === Bioroebe::Taxonomy.save_into_which_file?
  # ========================================================================= #
  def self.save_into_which_file?
    TAXONOMY_NCBI_DATABASE_LAST_UPDATE_LOG_FILE
  end

  # ========================================================================= #
  # === Taxonomy.save_when_we_last_updated_the_database
  #
  # Save into a file when we last updated the database. This method will
  # make use of the constant SAVE_FILE, which tells us where to keep
  # the save file, and can be found at: lib/taxonomy/shared/shared.rb
  # ========================================================================= #
  def self.save_when_we_last_updated_the_database(
      be_verbose = ::Bioroebe::BE_VERBOSE
    )
    if be_verbose # BE_VERBOSE is defined in shared/shared.rb
      e "Trying to log last update into file "\
        "#{::Bioroebe.sfile(save_into_which_file?)} next."
    end
    # ======================================================================= #
    # Only store into a file if the constant SHALL_WE_LOG_LAST_UPDATE
    # is set to true.
    # ======================================================================= #
    if SHALL_WE_LOG_LAST_UPDATE
      into = save_into_which_file?
      # ===================================================================== #
      # Verify that we can actually write into the directory.
      # ===================================================================== #
      if File.stat(File.dirname(into)).writable?
        ::Bioroebe.write_what_into(
          'Last Update of the Taxonomy NCBI Database: '+
          ::Bioroebe.return_current_day_month_year+', '+
          ::Bioroebe.return_current_hours_minutes_seconds+N,
          into
        )
      end
    end
  end

  # ========================================================================= #
  # === Bioroebe::Taxonomy.e
  # ========================================================================= #
  def self.e(i = '')
    ::Bioroebe.e(i)
  end

  # ========================================================================= #
  # === Bioroebe::Taxonomy.cd
  # ========================================================================= #
  def self.cd(i = '..')
    ::Bioroebe.cd(i)
  end

  # ========================================================================= #
  # === Taxonomy.ensure_that_temp_dir_exists
  # ========================================================================= #
  def self.ensure_that_temp_dir_exists
    ::Bioroebe.ensure_that_the_base_directories_exist
  end

  # ========================================================================= #
  # === Bioroebe.show_remote_urls_to_the_NCBI_taxonomy_webpage
  #
  # This method will show the remote URLs to different tax-IDs.
  #
  # If no argument is given (thus no ID) then the NCBI taxonomy
  # parts are displayed.
  # ========================================================================= #
  def self.show_remote_urls_to_the_NCBI_taxonomy_webpage(optional_id = nil)
    if optional_id
      erev 'http://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?mode=Info&id='+optional_id.to_s
    else
      erev 'The remote URLs towards the NCBI taxonomy part are:'
      e
      erev '  NCBI: '+sfancy('http://www.ncbi.nlm.nih.gov/taxonomy')
      erev '  NCBI: '+sfancy('http://www.ncbi.nlm.nih.gov/taxonomy?term=1')
      erev '  NCBI: '+sfancy('http://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi')
      e
    end
  end; self.instance_eval { alias browse show_remote_urls_to_the_NCBI_taxonomy_webpage } # === Bioroebe.browse

  require 'bioroebe/constants/urls.rb'
  # ========================================================================= #
  # === Bioroebe::Taxonomy.update                              (download tag)
  #
  # This class method will obtain the file taxdump.tar.gz.
  #
  # By default we will download into TEMP_DIR, which at the moment of
  # documenting this method defaults to /tmp/robert/ on the I.M.P.
  # cluster, or rather the value that is stored in the very constant
  # DOWNLOAD_DIR.
  # ========================================================================= #
  def self.update(
      this_dir   = ::Bioroebe.taxonomy_download_directory?,
      be_verbose = true
    )
    case be_verbose
    when :be_silent
      be_verbose = false
    end
    ::Bioroebe.show_time_now # Display the current start-time.
    ::Bioroebe.ensure_that_the_base_directories_exist
    unless File.directory? this_dir
      ::Bioroebe.mkdir(this_dir)
    end
    if Dir.exist? this_dir # All is fine here.
    else # Safeguard here.
      if at_home?
        if be_verbose
          this_dir = '/home/Temp/' # Hardcoded.
          e "The directory at #{sdir(this_dir)} does not "\
            "exist, thus trying to"
          e 'use '+sdir(this_dir)+' instead.'
        end
      else # Then we create that directory.
        ::Bioroebe.mkdir(this_dir) unless Dir.exist? this_dir
      end
    end
    if be_verbose
      e 'Now trying to change to the base directory at '+
        ::Bioroebe.sdir(this_dir)+'.'
    end
    cd this_dir
    e N+'We will next attempt to download the file `'+
      ::Bioroebe.sfancy(URL_TO_TAXONOMY_ARCHIVE)+'`'
    e 'into the local directory `'+
       ::Bioroebe.sdir(this_dir)+'` via wget.'
    if be_verbose
      e '(This file will usually have a file size of about '\
        '~36 MB. [Last verification of this size: Dec 2016])'
    end
    local_file = File.basename(URL_TO_TAXONOMY_ARCHIVE)
    ::Bioroebe.remove_file(local_file) if File.exist? local_file
    ::Bioroebe.wget_download(URL_TO_TAXONOMY_ARCHIVE) # Use class Download to download something.
    ::Bioroebe.extract(local_file)
    save_when_we_last_updated_the_database
  end; self.instance_eval { alias download_ncbi_database update } # === Bioroebe::Taxonomy.download_ncbi_database

  # ========================================================================= #
  # === Bioroebe::Taxonomy.taxonomy_download_directory?
  #
  # This method will return the download directory for use in the
  # Taxonomy subcomponent.
  # ========================================================================= #
  def self.taxonomy_download_directory?
    "#{::Bioroebe.log_dir?}taxonomy/"
  end

end

# =========================================================================== #
# === Bioroebe.taxonomy_download_directory?
# =========================================================================== #
def self.taxonomy_download_directory?
  ::Bioroebe::Taxonomy.taxonomy_download_directory?
end

end

if __FILE__ == $PROGRAM_NAME
  puts Bioroebe::Taxonomy.taxonomy_download_directory?
  puts Bioroebe::Taxonomy.update  
end # taxonomydownloaddirectory