#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Bioroebe::RNALfoldWrapper
#
# This is just a custom class that can be used to invoke the program
# RNALfold, hence the name of this class being RNALfoldWrapper.
#
# It's usefulness will be very limited though.
# =========================================================================== #
require 'bioroebe/base/commandline_application/commandline_application.rb'

module Bioroebe

class RNALfoldWrapper < ::Bioroebe::CommandlineApplication # === Bioroebe::RNALfoldWrapper

  # ========================================================================= #
  # === NAMESPACE
  # ========================================================================= #
  NAMESPACE = inspect

  # ========================================================================= #
  # === START_VALUE
  #
  # This is the start value for L.
  # ========================================================================= #
  START_VALUE = 150

  # ========================================================================= #
  # === UP_TO
  # ========================================================================= #
  UP_TO = 500

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_arguments = ARGV,
      run_already           = true
    )
    reset
    set_commandline_arguments(
      commandline_arguments
    )
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    super()
    # ======================================================================= #
    # === @namespace
    # ======================================================================= #
    @namespace = NAMESPACE
    # ======================================================================= #
    # === @target_directory
    #
    # Specify which directory is to be used. This can be toggled between
    # the two major names, e. g. 'cdna/' or 'ncrna/'.
    # ======================================================================= #
    @target_directory = 'cdna'
    @target_directory = 'ncrna'
    # ======================================================================= #
    # === @use_zscore
    # ======================================================================= #
    @use_zscore = true
    # ======================================================================= #
    # === @use_this_zscore_value
    # ======================================================================= #
    @use_this_zscore_value = -2.5 # This is the default value.
    # ======================================================================= #
    # === @use_this_chromosome_number
    #
    # Which chromosome number to use, e. g. 8 or 15 in my case.
    # ======================================================================= #
    @use_this_chromosome_number = 15 # 8
    # ======================================================================= #
    # === @array_cmds
    # ======================================================================= #
    @array_cmds = []
    # ======================================================================= #
    # === @also_run_these_commands
    #
    # If the next variable is set to true then we will also run these
    # commands, after having shown them first. By default this is not
    # the case.
    # ======================================================================= #
    @also_run_these_commands = false
  end

  # ========================================================================= #
  # === use_this_zscore_value?
  # ========================================================================= #
  def use_this_zscore_value?
    @use_this_zscore_value
  end; alias zscore_value? use_this_zscore_value? # === zscore_value?

  # ========================================================================= #
  # === menu
  # ========================================================================= #
  def menu(
      i = @commandline_arguments
    )
    if i.is_a? Array
      i.each {|entry| menu(entry) }
    else
      case i # case tag
      # ===================================================================== #
      # === rnafoldwrapper --run
      # ===================================================================== #
      when /^-?-?run$/
        @also_run_these_commands = true
      end
    end
  end

  # ========================================================================= #
  # === array_cmds?
  # ========================================================================= #
  def array_cmds?
    @array_cmds
  end; alias main_array? array_cmds? # === main_array?

  # ========================================================================= #
  # === use_zscore?
  # ========================================================================= #
  def use_zscore?
    @use_zscore
  end

  # ========================================================================= #
  # === report_where_the_commands_will_be_stored_and_store_them
  # ========================================================================= #
  def report_where_the_commands_will_be_stored_and_store_them
    _ = determine_log_filename
    opnn; e "These commands will be stored in the file `#{sfile(_)}`."
    what = main_array?.join(N)
    write_what_into(what, _)
  end

  # ========================================================================= #
  # === determine_log_filename
  #
  # This method should return the log filename.
  # ========================================================================= #
  def determine_log_filename
    _ = @target_directory.dup
    if use_zscore?
      _ << "_zscore_value_of_#{zscore_value?}"
    end
    _ << '.log'
    _
  end

  # ========================================================================= #
  # === run
  # ========================================================================= #
  def run
    menu
    # ======================================================================= #
    # Increment from 150 to 500 normally.
    # The line should be e. g.:
    #   RNALfold -L 200 < cdna/cdna.8.fa > cdna/cdna.8.fa.200.lfold
    # ======================================================================= #
    START_VALUE.step(UP_TO, 50) {|index|
      cmd_to_run = "RNALfold ".dup
      if @use_zscore
        cmd_to_run << "--zscore=#{@use_this_zscore_value} "
      end
      cmd_to_run << "-L #{index} < "\
                    "#{@target_directory}/#{@target_directory}.#{@use_this_chromosome_number}.fa "\
                    "> "
      name_of_the_output_file = "#{@target_directory}/#{@target_directory}.#{@use_this_chromosome_number}.fa.".dup
      if use_zscore?
        name_of_the_output_file << "zscore_value_of_#{@use_this_zscore_value}___"
      end
      name_of_the_output_file << "L-value-used-was___#{index}.lfold"
      cmd_to_run << name_of_the_output_file
      e cmd_to_run
      @array_cmds << cmd_to_run
    }
    report_where_the_commands_will_be_stored_and_store_them
    if @also_run_these_commands
      opnn; e 'These commands will also be run next.'
      @array_cmds.each {|entry| esystem entry }
    end
  end

end; end

if __FILE__ == $PROGRAM_NAME
  Bioroebe::RNALfoldWrapper.new
end # rnafoldwrapper