#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'bioroebe/constants/aminoacids_and_proteins.rb'
# =========================================================================== #
module Bioroebe

  # ========================================================================= #
  # === ARRAY_AMINOACIDS_THAT_CAN_BE_PHOSPHORYLATED
  #
  # Just list the aminoacids that can typically be phosphorylated.
  # ========================================================================= #
  ARRAY_AMINOACIDS_THAT_CAN_BE_PHOSPHORYLATED = %w(
    S Y T
  )

  # ========================================================================= #
  # === ENGLISH_LONG_NAMES_FOR_THE_AMINO_ACIDS
  #
  # We have to keep the long names for the amino acids in one
  # constant, so that we can do queries lateron.
  # ========================================================================= #
  ENGLISH_LONG_NAMES_FOR_THE_AMINO_ACIDS = (%w( 
    Alanine
    Arginine
    Asparagine
    Cysteine
    Glutamine
    Glycine
    Histidine
    Isoleucine
    Leucine
    Lysine
    Methionine
    Phenylalanine
    Proline
    Serine
    Threonine
    Tryptophane
    Tyrosine
    Valine
  ) << 'Aspartic acid' << 'Glutamic acid').sort

  # ========================================================================= #
  # === POSSIBLE_AMINO_ACIDS
  #
  # Which Aminoacids are possible/allowed? We will list them here:
  #
  #   ACDEFGHIKLMNPQRSTUVWY
  #
  # Note that this is distinct from the constant AMINO_ACIDS, which is
  # instead loaded from a local .yml file. This constant includes
  # all the 20 canonical aminoacids, whereas AMINO_ACIDS may also
  # include pyrrolysine and selenocysteine.
  # ========================================================================= #
  POSSIBLE_AMINO_ACIDS = 'ACDEFGHIKLMNPQRSTVWY'
    TWENTY_CANONICAL_AMINOACIDS = POSSIBLE_AMINO_ACIDS # === TWENTY_CANONICAL_AMINOACIDS

  # ========================================================================= #
  # === ARRAY_AMINO_ACIDS_ALPHABET
  #
  # This keeps an Array with all aminoacids, in one-letter format.
  #
  # So it is equivalent to:
  #
  #   ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"]
  #
  # ========================================================================= #
  ARRAY_AMINO_ACIDS_ALPHABET = POSSIBLE_AMINO_ACIDS.chars

  # ========================================================================= #
  # === Bioroebe.all_aminoacids?
  #
  # This method will return all available aminoacids.
  #
  # Example:
  #
  #   Bioroebe.all_aminoacids? # => ["A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "P", "Q", "R", "S", "T", "V", "W", "Y"]
  #
  # ========================================================================= #
  def self.all_aminoacids?
    ARRAY_AMINO_ACIDS_ALPHABET
  end

  # ========================================================================= #
  # === Bioroebe.filter_away_invalid_aminoacids
  #
  # Usage example:
  #
  #   Bioroebe.filter_away_invalid_aminoacids('ATMÜ') # => "ATM"
  #
  # ========================================================================= #
  def self.filter_away_invalid_aminoacids(i)
    array_that_is_allowed = all_aminoacids?
    return i.chars.select {|entry| array_that_is_allowed.include? entry }.join
  end

  # ========================================================================= #
  # === all_aminoacids?
  # ========================================================================= #
  def all_aminoacids?
    ::Bioroebe.all_aminoacids?
  end

  # ========================================================================= #
  # === AMINOACID_FAMILIES
  # ========================================================================= #
  AMINOACID_FAMILIES = {
    'citratzyklus' => {
      # Alpha-Ketoglutarat: EPQR
      'alpha-ketoglutarat' => %w( E P Q R ),
      # Oxalacetat: DMN-KTI
      'oxalacetat' => %w( D N K M T I ),
    },
    'glykolyse' => {
      'pyruvat' => %w( A V L ),                 # AVL
      '3-phosphoglycerinsäure' => %w( S G C ), # SGC
      },
      'chorismat' => {
        'aromatische_familie' => %w( F Y W )       # FYW
      },
      'ribose-5-p' => {
        'histidinol' => %w( H ) # Histidine.
      },
  }

  # ========================================================================= #
  # === Bioroebe.aminoacid_families?
  #
  # Feedback which aminoacid-families we know of.
  #
  # Usage example:
  #
  #   pp Bioroebe.aminoacid_families?; ''
  #
  # ========================================================================= #
  def self.aminoacid_families?
    AMINOACID_FAMILIES
  end; self.instance_eval { alias aa_families? aminoacid_families? } # === Bioroebe.aa_families?

end

if __FILE__ == $PROGRAM_NAME
  pp Bioroebe.aa_families?
end